import numpy as np
from faps.paternityArray import paternityArray
from pandas import read_csv

def read_paternity_array(path, mothers_col=1, likelihood_col = 2, fathers_col=None, delimiter=","):
    """
    Import a text file containing of log likelihoods of paternity.

    This is a shortcut to save having to run paternity_array every time you open
    a data set. Note that only information on the IDs of the offspring and
    mothers, likelihoods of paternity for each candidate and likelihoods that
    fathers are missing can be retrieved; other information; for example
    covariates must be generated again.

    Parameters
    ----------
    path: str
        filename path to locate the text file. This should contain unique ID
        labels for each offspring individual in the first column, followed by
        (optional)columns of names for the mothers and fathers. After this
        include log likelihoods that each adult is the sire of each offspring.
        The header row should include names for each candidate, or else be left
        blank. The final column should be a vector of likelihoods that the true
        sire of each offspring has not been sampled.
    mothers_col: int
        If a column of maternal names has been included, indicate column index
        here. This is usually in the second column, which in Python speak is
        column 1.
    likelihood_col: int
        indicate the column index where genotype information begins. This is
        usually in the third column, which in Python speak is column 2, unless
        you have given a column for fathers.
    delimiter: str
        The symbol used to separate values in the text file. Defaults to
        commas.

    Returns:
    --------
    A paternityArray object with attributes for likehood of paternity,
    likelihoods that father are missing, IDs for the offspring and mothers,
    and a vector of NAs for the fathers. Other possible attributes of a
    paternityArray are set to `None`.

    Examples
    --------
    from faps import *
    import numpy as np

    # Generate a population of adults
    allele_freqs = np.random.uniform(0.3,0.5,50)
    adults = make_parents(20, allele_freqs)
    # Mate the first adult to the next three.
    mother = adults.subset(0)
    progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
    # Create paternityArray
    patlik = paternity_array(progeny, mother, adults, mu=0.0013)

    # Write the paternityArray object to disk
    patlik.write("my_paternity_array.csv")
    # Read it back in
    patlik2 = read_paternity_array("my_paternity_array.csv")
    # Split the array into paternal families
    patlik2.split(by=progeny.fathers)
    """
    # Import data frame.
    df = read_csv(path, delimiter=delimiter)

    # Create a new paternityArray object.
    pat_array = paternityArray(
        likelihood = df[df.columns[likelihood_col:-1]].to_numpy(),
        lik_absent = df[df.columns[-1]].to_numpy(),
        offspring  = df[df.columns[0]].to_numpy(),
        mothers    = df[df.columns[mothers_col]].to_numpy(),
        fathers    = np.repeat('NA', len(df)),
        candidates = df.columns[likelihood_col:-1]
    )

    return pat_array



import numpy as np
from faps.genotypeArray import genotypeArray

def accuracy_paternity(progeny, parents, paternity_array, partition, target_parent = 'father'):
    """
    Summarise how well a sibship partition returns information on the posterior
    probability of paternity of true sires. For individuals whose fathers have not been
    sampled, the probability that the father is drawn from population allele frequencies
    is returned.

    Deprecated in favour of a function within sibshipCluster.

    Parameters
    ----------
    progeny: genotypeArray
        Genotype information on offspring individuals included in paternity_array.

    parents: genotypeArray
        Genotype information on the candidate parents included in paternity_array.

    paternity_array: A matrix of probabilities that each candidate parent is the true
        parent of each individual. This array can be generated with paternity_array, or
        else supplied independently. Rows show index offspring, and columns parents.

    target_parent: String denoting whether the paternity_array indexes fathers, mothers,
        or parents pairs. Valid arguments are 'mother', 'father' and 'parents',  or
            equivalently 'm' and 'f' respectively.

    Returns
    -------
    A list of two elements:
    1. The mean posterior probability of sampled true sires, conditioned on the father
        having been included in the sampled of candidate parents.
    2. The mean posterior probability that the sire is drawn from population allele
        frequencies, conditioned on the true sire not having been included in the sampled
        of candidate parents.
    """
    # PROBABILITIES OF PATERNITY FOR THE TRUE SIRES
    # Check which individuals still have their father in the parent array, and which do not.
    if target_parent is 'mother' or target_parent is 'm':
        parent_present = [progeny.mothers[x]     in parents.names for x in range(progeny.size)]
        parent_missing = [progeny.mothers[x] not in parents.names for x in range(progeny.size)]
    elif target_parent is 'father' or target_parent is 'f':
        parent_present = [progeny.fathers[x]     in parents.names for x in range(progeny.size)]
        parent_missing = [progeny.fathers[x] not in parents.names for x in range(progeny.size)]
    elif target_parent is 'parent' or target_parent is 'p':
        parent_present = [progeny.parents[x]     in parents.names for x in range(progeny.size)]
        parent_missing = [progeny.parents[x] not in parents.names for x in range(progeny.size)]
    else:# target_parent not in ['mother', 'father', 'parent', 'm', 'f', 'p']:
        raise ValueError("target_parent must be 'mother', 'father', or 'parents'.\nThis should be whichever is indexed in the columns of paternity_array.")

    # If there are no individuals with fathers present, return
    if sum(parent_present) == 0:
        mean_patprob = np.nan
    if sum(parent_present) >= 1:
        # subset the progeny to select just those non-orphaned offspring
        not_orphans = progeny.subset(np.arange(progeny.size)[np.array(parent_present)])
        # get the positions of the true fathers in the probability matrix for these individuals.
        father_pos  = not_orphans.parent_index(target_parent, parents.names)
        # Get the posterior probability of the true sire of each individual within its FSG
        # for those individuals with a father in the array.
        pat_probs = np.zeros(sum(parent_present)) # empty vector to store probabilities
        for x in range(sum(parent_present)):
            offs_pos     = np.arange(progeny.size)[np.array(parent_present)][x] # position of the offspring
            # paternity probabilities multiplied across individuals in the sibship
            this_gamma   = paternity_array[partition == partition[offs_pos]].sum(0)
            this_gamma   = this_gamma - alogsumexp(this_gamma) # normalise to sum to one.
            # get the value for the true sire
            pat_probs[x] = this_gamma[father_pos[x]]
        mean_patprob = pat_probs.mean()

    # POSTERIOR PROBABILITIES THAT A SIRE WAS NOT SAMPLED.
    # If there are no individuals with missing fathers, return NA
    if sum(parent_missing) == 0:
        mean_absprob = np.nan
    # Otherwise, get the posterior probabilities that each sire is missing
    if sum(parent_missing) >= 1:
        # subset the progeny to select just those orphaned offspring
        orphans     = progeny.subset(np.arange(progeny.size)[np.array(parent_missing)])
        # Get the posterior probability that the true sire is missing for each individual
        abs_probs   = np.zeros(sum(parent_missing)) # empty vector to store probabilities
        for x in range(sum(parent_missing)):
            offs_pos     = np.arange(progeny.size)[np.array(parent_missing)][x] # position of the offspring
             # paternity probabilities for this sibship
            this_gamma   = paternity_array[partition == partition[offs_pos]].sum(0)
            this_gamma   = this_gamma - alogsumexp(this_gamma) # normalise to sum to one.
            # The probability of an unsampled father is given in the final column
            abs_probs[x] = this_gamma[-1]
        mean_absprob = abs_probs.mean()

    return [mean_patprob, mean_absprob]



def calculate_SNP_calls(geno_probs):
    """
    Convert genotype probabilities to hard calls.

    Probabilities in geno_probs.geno_probs are rounded to either
    zero or one, meaning that whichever genotype has probability >0.5
    is called. If all probabilities are < 0.5, no SNP is called
    (returns np.nan)

    Parameters
    ----------
    geno_probs: array
        3-dimensional array indexing individuals, loci, and three
        possible genotypes. This is the same format as the
        `geno_probs` attribute of a genotypeArray object.

    Returns
    -------
    3-dimensional array indexing individuals, loci, and two chromosomes.
    This is the same format as the `geno` attribute of a
    genotypeArray object.

    Example
    -------

    """
    # Round genotype probabilities to zero or one.
    call_probs = np.round(geno_probs.geno_probs)
    # 3-dimensional array of np.nan
    geno = np.full([geno_probs.size, geno_probs.nloci,2], np.nan)
    # Fill in entries called as zero
    geno[:,:,0][call_probs[:,:,0] == 1] = 0
    geno[:,:,1][call_probs[:,:,0] == 1] = 0
    # Fill in entries called as heterozygote
    geno[:,:,0][call_probs[:,:,1] == 1] = 0
    geno[:,:,1][call_probs[:,:,1] == 1] = 1
    # Fill in entries called as 2.
    geno[:,:,0][call_probs[:,:,2] == 1] = 1
    geno[:,:,1][call_probs[:,:,2] == 1] = 1

    return(geno)



import numpy as np
from faps.genotypeArray import genotypeArray
from faps.calculate_geno_probs import calculate_geno_probs

def make_parents(size, allele_freqs, mu=1e-12, family_name = 'base'):
    """
    Draw a base population of reproductive individuals from population allele frequencies.

    Parameters
    ----------
    size: integer
        Number of individuals to create.
    allele_freqs: array-like
        Vector of allele frequencies.
    mu: float or 1-d array between 0 and 1
        Per locus genotype error rate; the probability that the called
        genotype is incorrect. Alternatively, supply a vector of error rates
        for each locus. Defaults to 1e-12.
    family_name: str, optional
        String denoting the name for this family.

    Returns
    -------
    A genotypeArray object.
    """
    if size < 2 or not isinstance(size, int):
        raise ValueError("Size must be an integer of 2 or more.")
    if np.any(allele_freqs > 1) or np.any(allele_freqs < 0):
        raise ValueError("One or more allele frequency values exceeds zero or one.")
    else:
        # draw alleles for each individual.
        genomes = [np.reshape(np.random.binomial(1,p,2*size), [2,size]) for p in allele_freqs]
        genomes = np.array(genomes).astype(float)
        genomes = np.rollaxis(genomes, 2, 0) # reorder the array.

        offspring_names   = np.array([family_name + '_' + str(a) for a in np.arange(size)])
        maternal_names    = np.repeat('NA', size)
        paternal_names    = np.repeat('NA', size)

        geno_probs = calculate_geno_probs(genomes, mu)

        geno = genotypeArray(
            geno    = genomes,
            geno_probs = geno_probs,
            names   = offspring_names,
            mothers = maternal_names,
            fathers = paternal_names,
            markers = np.arange(len(allele_freqs))
        )
        return geno



import pandas as pd
import numpy as np
from faps.sibshipCluster import sibshipCluster


def summarise_sires(sibships, drop_na=False):
    """
    Summarise mating events across maternal families.

    This is a wrapper function to call `sibshipCluster.sires()` across
    multiple sibshipCluster objects stored in a dictionary. It is assumed that
    each object represents a maternal family.

    Parameters
    ----------
    sibships: dict
        Dictionary of sibshipCluster objects for multiple maternal families.
        Keys should normally indicate maternal family name.

    Returns
    -------
    DataFrame giving mother (taken from the keys of the input dictionary),
    fathers (inherited from each sibshipCluster object), probabilties of
    having sired at least one offspring, and the expected number of offspring.

    Examples
    --------
    from faps import *
    import numpy as np

    # Generate a population of adults
    allele_freqs = np.random.uniform(0.3,0.5,50)
    adults = make_parents(20, allele_freqs)

    # Mate the first adult to the next three.
    mother = adults.subset(0)
    progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
    patlik = paternity_array(progeny, mother, adults, mu=0.0013)
    # Split offspring by paternal family
    patlik = patlik.split(by=progeny.fathers)
    # Cluster each paternal family into sibships
    sc = sibship_clustering(patlik)

    summarise_sires(sc)

    """

    if not isinstance(sibships, dict):
        raise TypeError("`sibships` should be a dictionary of sibshipCluster onjects.")
    if not all([isinstance(x, sibshipCluster) for x in sibships.values()]):
        raise TypeError("Not all elements of `sibships` are sibshipCluster objects.")
    if isinstance(sibships, sibshipCluster):
        raise TypeError("`summarise_sires` is intended to work on a dictionary of sibshipCluster objects, but a single sibshipCluster object was supplied. In this case, call `sires()` directly on the onject, i.e. object.sires().")

    # For each mother, get a list of plausible mates
    # with the probability of each having sired at least one offspring.
    me_dict = {k: v.sires() for k,v in sibships.items()}
    # Get distances to those sires for each maternal family.
    # Return a list of dataframes for each mother
    me_df = pd.concat(list(me_dict.values()))
    me_df = []
    for k in me_dict.keys():
        df = pd.DataFrame({
            'mother'   : k,
            'father'   : me_dict[k]['label'],
            'log_prob' : me_dict[k]['log_prob'],
            'prob'     : me_dict[k]['prob'],
            'offspring': me_dict[k]['offspring']
        })
        me_df = me_df + [df]
    # Coerce list into a single data frame listing mother, father, distance and log probability
    me_df = pd.concat(me_df)
    me_df = me_df.reset_index(drop=True)
    return me_df



import numpy as np
from faps.pr_transition import pr_transition
from faps.genotypeArray import genotypeArray

def lik_sampled_fathers(offspring, mothers, males, mu, mother_index=None):
    """
    Calculate a matrix of multilocus transition probabilities for all possible triplets
    of offspring, their known mothers, and a set of candidate males. This sums over all
    possible offspring and parental genotypes.

    Parameters
    ----------
    offspring: genotypeArray
        Array of observed genotype data for the offspring.

    mothers: genotypeArray
        An array of observed genotype data for the mothers. Data on mothers need
        to be in the same order as those for the offspring. If a whole parent
        array is used, this can be subsetted into the correct order using the
        option mother_index.

    fathers: genotypeArray
        Array of observed genotype data for the candidate males.

    mu: float
        Point estimate of the genotyping error rate. Note that sibship clustering
        is unstable if mu_input is set to exactly zero. Any zero values will
        therefore be set to a very small number close to zero (10^-12).

    mother_index: list, optional
        List of integers indexing the position of the mother of
        each offspring in the maternal_genotypes array. If left blank, the whole array
        is used.

    Returns
    -------
    A matrix of log likelihoods of paternity, with a row for each offspring and a
    column for each candidate male.
    """
    if not isinstance(offspring, genotypeArray):
        raise TypeError('offspring data not given as a genotypeArray.')
    elif not isinstance(mothers, genotypeArray):
        raise TypeError('mothers not given as a genotypeArray.')
    elif not isinstance(males, genotypeArray):
        raise TypeError('male genotype data not given as a genotypeArray.')

    # If a list of indices for the mothers has been given, subset the maternal data
    if mother_index is not None:
        mothers = mothers.subset(mother_index)
    # if mu is given as zero, set this to a very small number for sibship clustering
    if mu == 0: mu = 1.0 / 10**12

    # Diploid genotypes of each dataset.
    male_diploid = males.geno.sum(2)
    moth_diploid = mothers.geno.sum(2)
    offs_diploid = offspring.geno.sum(2)

    # empty array to store probabilities.
    lik_trans = np.zeros([offspring.size, males.size, males.nloci])
    # loop over all possible gemotype combinations and get transitions probs for each.
    for o in [0,1,2]:
        for m in [0,1,2]:
            for f in [0,1,2]:
                lik_trans += pr_transition(offs_diploid, moth_diploid, male_diploid, o, m, f, mu)
    lik_trans = lik_trans # correct for the number of combinations.
    with np.errstate(divide='ignore'): lik_trans = np.log(lik_trans) # convert to log.

    # sum diploid genotypes to identify where the negative values (i.e. dropouts) are.
    dropout_mask = offs_diploid[:, np.newaxis] + moth_diploid[:, np.newaxis] + male_diploid[np.newaxis]
    # array of the number of loci for each trio for which one or more individual had a dropout.
    valid_loci   = males.nloci - (dropout_mask < 0).sum(2)

    # Correct for dropouts
    lik_trans[dropout_mask < 0] = 0 # set loci with one or more dropouts to zero to allow summing.
    lik_trans = lik_trans.sum(2) # sum over loci.
    lik_trans = lik_trans * (float(males.nloci) / valid_loci) # scale by the number of valid SNPs.

    return lik_trans



import numpy as np

def convert_genotypes(diploid_data):
    """
    Convert a table of diploid genotypes to an array of haploid genotypes for a
    genotypeArray object. Usually, this will be an internal function call when
    importing genotype data from a text file to a genotypeArray.

    This function copies diploid data to each of two columns for haploid genotypes.
    Heterozygotes are assigned a 0 for locus 1 and a 1 for locus 2.
    Major allele homozygotes are assigned as 1 for both loci.
    Minor allele homozygotes are already coded 0 at both loci.

    Parameters
    ----------
    diploid_data: array
        Array containing diploid genotypes of candiate parents, with a row for
        every individual and a column for every SNP. The first column should
        label maternal IDs, or else left blank. The first row should label SNPs,
        or else be left blank.

    Returns
    -------
    Haploid genotype array for genotypeArray objects.
    """
    n_inds = diploid_data.shape[0]         # number of individuals in the array
    n_loci = diploid_data.shape[1]         # number of loci, accounting for the index column.

    # parMat = np.zeros([n_inds,n_loci,2]).astype(float)    # empty matrix to store data.
    parMat = np.full([n_inds,n_loci,2], np.nan)

    # First allele
    parMat[:,:,0][diploid_data == 0] = 0 # For 0/0 homozygotes, set to 0
    parMat[:,:,0][diploid_data == 1] = 0 # for heterozygotes, set this locus to 0.
    parMat[:,:,0][diploid_data == 2] = 1 # for minor allele homozygotes, set this locus to 1.
    # Second allele
    parMat[:,:,1][diploid_data == 0] = 0 # For 0/0 homozygotes, set to 0
    parMat[:,:,1][diploid_data == 1] = 1 # for heterozygotes, set this locus to 0.
    parMat[:,:,1][diploid_data == 2] = 1 # for minor allele homozygotes, set this locus to 1.

    # parMat[:,:,1]                     = np.copy(diploid_data)
    # parMat[:,:,1][parMat[:,:,1] == 2] = 1       # for minor allele homozygotes, set this locus to 1.

    return parMat



import numpy as np

def relation_matrix(partitions):
    """
    Create a matrix of pairwise relationships within a half-sibling array
    given a real or hypothesised vector of partitions.
    Relationships are coded as 1 if they are full siblings and 0 if not.
    """
    sibshipsize = len(partitions) # total offspring number.
    relmat      = np.zeros([sibshipsize, sibshipsize]) # empty matrix of zeros.
    siblabel    = np.unique(partitions) # vector of unique labels for each full sib family.
    index       = np.arange(sibshipsize) # indices for each sib
    
    # for each full sibship fill in the pairwise elements with ones.
    for i in range(len(siblabel)):
        b                        = partitions == siblabel[i] # truth array
        rowvals                  = np.repeat(index[b], sum(b))  
        colvals                  = np.tile(  index[b], sum(b))
        relmat[rowvals, colvals] = 1
    
    return relmat


import numpy as np
from faps.convert_genotypes import convert_genotypes
from faps.genotypeArray import genotypeArray
from faps.calculate_geno_probs import calculate_geno_probs

def read_genotypes(path, genotype_col = 1, mothers_col=None, fathers_col=None, mu=1e-12, delimiter=","):
    """
    Import a text file containing of diploid SNP data to FAPS.

    Retrieves the file containing genotype data, and convert this to a genotypeArray.
    If the data are for offspring individuals, optional columns can be used to indicate
    their labels to match up with a file of parental data. Otherwise, names for mothers
    and fathers are given as 'NA' in the output genotypeArray.

    Parameters
    ----------
    path: str
        Filename path to locate the text file. This should contain unique ID
        labels for each individual in the first column, followed by (optional)
        columns of names for the mothers and fathers. After this include diploid
        genotype information for each locus, with a single column for each marker.
        The header row should include names for each locus, or else be left blank.
        Homozygotes should be labelled 0 or 2, and heterozygotes as 1. Missing
        data should be labelled as 'NA' or -9 (negative nine).
    genotype_col: int
        indicate the column index where genotype information begins.
    mothers_col: int
        If a column of maternal names has been included, indicate column index
        here.
    fathers_col: If a column of paternal names has been included, indicate column
        index here.
    mu: float or 1-d array between 0 and 1
        Per locus genotype error rate; the probability that the called
        genotype is incorrect. Alternatively, supply a vector of error rates
        for each locus. Defaults to 1e-12.
    delimiter The symbol used to separate values in the text file. Defaults to
        commas.

    Returns
    -------
    A genotypeArray object.
    """
    # Import genotype data
    geno = np.genfromtxt(path, dtype='float', delimiter=delimiter)[1:, genotype_col:]
    # geno[geno < 0] = np.nan # set missing data to NA
    geno = convert_genotypes(geno)
    # pull out marker names.
    marker_names = np.genfromtxt(path, dtype=str)[0].split(",")[genotype_col:]

    # Import individual names
    names = np.genfromtxt(path, dtype='str')
    names = [names[i].split(",")[0].replace('"', '') for i in range(1, len(names))]

    # Names for the mothers, if these are given
    if isinstance(mothers_col, int):
        mothers = np.genfromtxt(path, dtype='str')
        mothers = [mothers[i].split(",")[mothers_col].replace('"', '') for i in range(1, len(mothers))]
    # If they aren't, return a list of NAs
    if mothers_col is None:
        mothers = np.repeat('NA', geno.shape[0])

    # Names for the fathers, if given
    if isinstance(fathers_col, int):
        fathers = np.genfromtxt(path, dtype='str')
        fathers = [fathers[i].split(",")[fathers_col].replace('"', '') for i in range(1, len(fathers))]
    # If they aren't, return a list of NAs
    if fathers_col is None:
        fathers = np.repeat('NA', geno.shape[0])

    geno_probs = calculate_geno_probs(geno, mu)

    output = genotypeArray(
        geno = geno,
        geno_probs = geno_probs,
        names = np.array(names),
        mothers = np.array(mothers),
        fathers = np.array(fathers),
        markers=np.array(marker_names)
        )
    return output



import numpy as np

def pr_unsampled(offspring_diploid, maternal_diploid, allele_freqs, offspring_genotype, maternal_genotype, male_genotype, mu):
    """
    Calculate the transitions probability for a given set of parental and offspring 
    alleles.

    Transitipn probabilities are then weight by the probability of drawing the allele
    from the population, and the probability that this allele is the true allele, given
    observed genotype data and the error rate mu.
    
    ARGUMENTS:
    offspring_diploid, maternal_diploid, male_diploid: arrays of diploid genotypes for
        the offspring, mothers and fathers.
    
    allele_freqs = vector of population allele frequencies.
    
    offspring_genotype, maternal_genotype, male_genotype: a two-element list of zeroes
        and ones indicating the diploid genotype of males, mothers and offspring to be
        considered.
    
    mu: point estimate of the genotyping error rate.
        
    RETURNS:
    A 3-dimensional array of probabilities indexing offspring, candidate males, and loci.
    These are given in linear, rather than log space.
    """
    # an array of all possible transition probabilities indexed as [offspring, mother, father].
    trans_prob_array = np.array([[[1,  0.5, 0  ],
                                  [0.5,0.25,0  ],
                                  [0,  0,   0  ]],
                                 [[0,  0.5, 1  ],
                                  [0.5,0.5, 0.5],
                                  [1,  0.5, 0  ]],
                                 [[0,  0,   0  ],
                                  [0,  0.25,0.5],
                                  [0,  0.5, 1  ]]])
    # the transition probability for the given genotypes.
    trans_prob = trans_prob_array[offspring_genotype, maternal_genotype, male_genotype]
    
    
    # Probabilities that the observed offspring marker data match observed data.
    pr_offs = np.zeros([offspring_diploid.shape[0], offspring_diploid.shape[1]])
    pr_offs[offspring_diploid == offspring_genotype] = 1-mu
    pr_offs[offspring_diploid != offspring_genotype] = mu
    
    # Probabilities that the observed maternal marker data match observed data.
    pr_mothers = np.zeros([maternal_diploid.shape[0], maternal_diploid.shape[1]])
    pr_mothers[maternal_diploid == maternal_genotype] = 1-mu
    pr_mothers[maternal_diploid != maternal_genotype] = mu
    
    # Probability of the father is drawn from population allele frequencies.
    if male_genotype is 0: pr_males = allele_freqs**2
    if male_genotype is 1: pr_males = allele_freqs*(1-allele_freqs)
    if male_genotype is 2: pr_males = (1-allele_freqs)**2

    return trans_prob * pr_males * pr_mothers * pr_offs


import numpy as np
from warnings import warn
from faps.alogsumexp import alogsumexp
from faps.unique_rows import unique_rows
from faps.squash_siblings import squash_siblings
from faps.paternityArray import paternityArray


def draw_fathers(partition, paternity_array=None, null_probs = None, ndraws=1000, use_covariates=False):
    """
    Draws a sample of compatible fathers for each family in a single partition.
    Candidates are drawn proportional to their posterior probability of paternity.

    Optionally, a sample of candidates can be drawn at random, or proportional
    to some other distribution, such as a function of distance.

    Parameters
    ----------
    partition: list
        A 1-d array of integers labelling individuals into families. This should
        have as many elements as there are individuals in paternity_probs.
    paternity_array: paternityArray
        Object listing information on paternity of individuals.
    null_probs: array, optional
        1-d vector of log probabilities that each candidate is the sire of each
        full sibship in the partition. Probabilities are assumed to be the same
        for each partition. If values do not sum to one, they will be normalised
        to do so.
    ndraws: int
        Number of Monte Carlo draws for each family.
    use_covariates: logical, optional
        If True, information on prbabilities associated with covariates stored
        in paternityArray objects are incorporated into weights for drawing likely
        fathers.

    Returns
    -------
    A list of candidates compatible with the genetic data, and a second list of
    candidates drawn under random mating if specified.
    """
    # number of sibships and compatible fathers
    nfamilies = len(np.unique(partition))

    if isinstance(paternity_array, paternityArray):
        if use_covariates is True:
            if paternity_array.covariate is 0:
                covar = 0
            elif isinstance(paternity_array.covariate, np.ndarray):
                if len(paternity_array.covariate.shape) > 1:
                    raise ValueError("covariate should be a 1-d array, but has shape {}".format(covariate.shape))
                if paternity_array.prob_array().shape[1] != len(paternity_array.covariate):
                    raise ValueError("Length of vector of covariates ({}) does not match the number of fathers ({})".format(len(paternity_array.candidates), paternity_array.covariate.shape[0]))
                if not all(paternity_array.covariate <= 0):
                    warn("Not all values in covariate are less or equal to zero. Is it possible probabilities have not been log transformed?")
                covar = paternity_array.covariate[np.newaxis]
            else:
                raise TypeError("If use_covariates is set to True, the covraiate attribute of the paternityArray object should be a 1-d array.")
        else:
            covar = 0

        nfathers  = len(paternity_array.candidates)+1
        # multiply likelihoods for individuals within each full sibship, then normalise rows to sum to 1.
        prob_array = squash_siblings(paternity_array.prob_array(), partition)
        prob_array = prob_array + covar
        prob_array = np.exp(prob_array - alogsumexp(prob_array,1)[:, np.newaxis])

    elif isinstance(null_probs, np.ndarray):
        if len(null_probs.shape) > 1:
            raise ValueError("null_probs is supplied should be a one dimensional vector for a single half-sibs array, but has shape {}.format(null_probs.shape}.")
        if not all(null_probs <= 0):
            warn("Not all values in null_probs are less or equal to zero. Is it possible probabilities have not been log transformed?")
        nfathers   = null_probs.shape[0]
        prob_array = null_probs - alogsumexp(null_probs)
        prob_array = np.tile(prob_array, nfamilies).reshape([nfamilies, len(null_probs)])
        prob_array = np.exp(prob_array)
    else:
        raise TypeError("Supply an array for either paternity_probs or null_probs.")
    if paternity_array is not None and null_probs is not None:
        warn('Values supplied for both paternity_probs and null_probs. null_probs will be ignored.')

    # generate a sample of possible paths through the matrix of candidate fathers.
    path_samples = np.array([np.random.choice(range(nfathers), ndraws, replace=True, p = prob_array[i]) for i in range(nfamilies)])
    path_samples = path_samples.T
    # identify samples with two or more famililies with shared paternity
    counts = [np.unique(path_samples[i], return_counts=True)[1] for i in range(len(path_samples))]
    valid  = [all((counts[i] == 1) & (counts[i] != nfathers))   for i in range(len(counts))]
    path_samples = np.array(path_samples)[np.array(valid)]
    output = [val for sublist in path_samples for val in sublist]

    return np.array(output)



import numpy as np
from faps import paternityArray

def paternity_list(offspring, mothers, males, allele_freqs, mu):
    """
    Calculate a matrix of paternity matrices for each of a set of half-sib arrays.

    This divides up offspring and maternal genotypeArrays into half-sib families
    based on information about the mother contained in the offspring genotypeArray.
    It then calculates a paternityArray for each, and returns these as a list.

    ARGUMENTS:
    offspring: A genotypeArray of observed genotype data for the offspring.

    mothers: an array of observed genotype data for the mothers. Data on mothers need
        to be in the same order as those for the offspring. If a whole parent array is
        used, this can be subsetted into the correct order using the option
        mother_index.

    males: A genotypeArray of observed genotype data for the candidate males.

    allele_freqs: vector of population allele frequencies for the parents.

    mu: point estimate of the genotyping error rate.

    RETURNS:
    A list of paternityArrays.
    """
    unique_families = np.unique(offspring.mothers)
    family_list     = []
    for i in range(len(unique_families)):
        this_family = offspring.subset(np.where(offspring.mothers == unique_families[i]))
        this_mother = mothers.subset(np.where(mothers.names == unique_families[[i]]))
        family_list+=  [paternityArray(this_family, this_mother, males, allele_freqs, mu)]
    return family_list



from pkg_resources import get_distribution
__version__ = get_distribution('faps').version

from faps.alogsumexp import alogsumexp
# data import and export
from faps.genotypeArray import genotypeArray
from faps.read_genotypes import read_genotypes
from faps.convert_genotypes import convert_genotypes
from faps.read_paternity_array import read_paternity_array
from faps.export_to_colony import export_to_colony
#from write_paternity_array import write_paternity_array
# Simulation
from faps.make_parents import make_parents
from faps.make_offspring import make_offspring
from faps.make_sibships import make_sibships
from faps.make_generation import make_generation
from faps.make_power import make_power
# paternity arrays
from faps.effective_nloci import effective_nloci
from faps.incompatibilities import incompatibilities
from faps.paternityArray import paternityArray
from faps.transition_probability import transition_probability
from faps.paternity_array import paternity_array
from faps.draw_fathers import draw_fathers
# clustering
from faps.pairwise_lik_fullsibs import pairwise_lik_fullsibs
from faps.sibshipCluster import sibshipCluster
from faps.lik_partition import lik_partition
from faps.sibship_clustering import sibship_clustering
from faps.summarise_sires import summarise_sires
# mating events
from faps.matingEvents import matingEvents
from faps.mating_events import mating_events
from faps.mating_summary import mating_summary



import numpy as np

def squash_siblings(paternity_probs, partitions):
    """
    Take an array of paternity likelihoods across fathers for each offspring, and multiply rows over putative siblings.
    
    Parameters
    ----------
    paternity_probs: array
        Array of posterior probabilities of paternity. Usually, this is the
        output of `prob_array` from a `paternityArray` object.
    
    partitions: array-like
        vector assigning a full sibship label to each individual.
    
    Returns
    -------
    An array with a row for every full sibship and a column for every candidate father.
    """
    partitions  = partitions.astype('int')
    labels      = np.unique(partitions)
    if len(labels) == 1:
        return paternity_probs.sum(0)[np.newaxis]
    else:
        fullsibliks = np.array([paternity_probs[partitions == f].sum(0) for f in labels])
        return fullsibliks




import numpy as np
from faps.genotypeArray import genotypeArray
from faps.calculate_geno_probs import calculate_geno_probs

def make_offspring(parents, noffs=None, dam_list=None, sire_list=None, mu=1e-12, family_name='offs'):
    """
    Mate individuals in a base population to create simulated offspring. Lists of
    specific sires and dams can be provided with the options dam_list and
    sire_list. If only the number of offspring are specified parents are mated at
    random from the base population.

    Parameters
    ----------
    parents: genotypeArray
        Genotype information on the parents to be mated.

    noffs: int
        Number of offspring to be produced. If specific dams and sires are
        specified, this is ignored.

    dam_list, sire_list: lists
        Integer lists of positions of sires and dams to be mated.
        Pairs are mated in order (i.e. the first dam with the first sire, and so
        forth). If used these two lists must be of the same length. If no
        arguments are given for either list, parents are mated at random with
        replacement, and the possibility of self-fertilisation.
    mu: float or 1-d array between 0 and 1
        Per locus genotype error rate; the probability that the called
        genotype is incorrect. Alternatively, supply a vector of error rates
        for each locus. Defaults to 1e-12.
    family_name: str, optional
        String denoting the name for this family.

    Returns
    -------
    A genotypeArray object.
    """
    if dam_list is None and sire_list is None and noffs is None:
        raise ValueError("Either noffs needs to be a positive integer, or else lists of dams and sires should be given.")

    # If parents haven't been specified, choose these at random.
    if dam_list is None and sire_list is None:
        if noffs < 1 or not isinstance(noffs, int):
            raise ValueError("noffs should be a positive integer.")
        nparents = parents.geno.shape[0]
        dam_list  = np.random.choice(range(nparents), noffs, replace=True).tolist()
        sire_list = np.random.choice(range(nparents), noffs, replace=True).tolist()
    # if parents have been specified, set noffs to the length of sires and dams.
    if dam_list is not None or sire_list is not None:
        noffs = len(dam_list)
        if len(dam_list) != len(sire_list):
            raise ValueError("List of sires must be the same length as the list of dams.")

    nloci = parents.geno.shape[1] # pull out the number of loci
    offs_genotypes= np.zeros([noffs, nloci, 2]) # empty array to store offspring genotypes.

    # pull out arrays of genotype data for the dams and sires.
    dam_genotypes  = parents.subset(dam_list).geno
    sire_genotypes = parents.subset(sire_list).geno

    # draw an array of indices for whether the first or second allele should be drawn.
    dam_alleles  = np.random.binomial(1, 0.5, nloci*noffs).reshape([noffs, nloci])
    sire_alleles = np.random.binomial(1, 0.5, nloci*noffs).reshape([noffs, nloci])
    # loop over every mating pair and send the selected alleles to offs_genotypes.
    for o in range(noffs):
        offs_genotypes[o,:,0] = np.array([dam_genotypes [o,l][dam_alleles [o,l]] for l in range(nloci)])
        offs_genotypes[o,:,1] = np.array([sire_genotypes[o,l][sire_alleles[o,l]] for l in range(nloci)])
    offs_genotypes = offs_genotypes.astype(float)

    # extra information on names.
    offspring_names   = np.array([family_name+'_'+str(a) for a in np.arange(noffs)])
    maternal_names    = parents.subset(dam_list).names
    paternal_names    = parents.subset(sire_list).names

    geno_probs = calculate_geno_probs(offs_genotypes, mu)

    return genotypeArray(
        geno = offs_genotypes,
        geno_probs = geno_probs,
        names = offspring_names,
        mothers = maternal_names,
        fathers = paternal_names,
        markers = np.arange(nloci)
    )



import numpy as np
from faps.genotypeArray import genotypeArray

def incompatibilities(parents, progeny):
    """
    Calculate the number of loci with Mendelian incompatibilities for arrays
    of offspring and candidate paternal genotypes.

    For each candidate-offspring dyad, this counts the number of loci at
    which the two individuals have opposing homozygous genotypes.

    Parameters
    ----------
    parents: genotypeArray
        Array of observed genotype data on the candidate fathers.
    progeny: genotypeArray
        Array of observed genotype data on the offspring.

    Returns
    -------
    An nxm array of integers, where n is the number of progeny and m is
    the number of candidate fathers.
    """
    # diploid genotypes of the parental and offspring diploid genotypes
    pdip = parents.geno.sum(2)
    odip = progeny.geno.sum(2)[:, np.newaxis]
    # Get number of incompatibilities for each offspring, for both pairs of possible incompatible genotypes
    arr1 = np.array([((pdip == 2) * (odip[i] == 0)).sum(1) for i in range(progeny.size)])
    arr2 = np.array([((pdip == 0) * (odip[i] == 2)).sum(1) for i in range(progeny.size)])
    # sum of incompatibilities for each possible combintation of incompatible genotypes
    return arr1 + arr2



import numpy as np
from faps.genotypeArray import genotypeArray
from faps.alogsumexp import alogsumexp

def sibship_paternity(offspring, mothers, males, allele_freqs, mu, partition, reciprocal = False):
    """
    Calculate the likelihoods that each candidate father is the sire of each
    full sibship implied by a partition sturcture. Includes likelihood that
    paternal alleles are drawn from population allele frequencies.


    Parameters
    ----------
    offspring: genotypeArray, or list of genotypeArrays
        Observed genotype data for the offspring.
    mothers: genotypeArray, or list of genotypeArrays
        Observed genotype data for the offspring. Data on mothers need
        to be in the same order as those for the offspring.
    males: genotypeArray
        Observed genotype data for the candidate males.
    allele_freqs: array-like
        Vector of population allele frequencies for the parents.
    mu: float between zero and one
        Point estimate of the genotyping error rate. Clustering is unstable if
        mu_input is set to exactly zero.
    partition: array
        1-d array of integers labelling individuals into families. This should
        have as many elements as there are individuals in paternity_probs.
    reciprocal: bool, optional
        If true, function return 1-transition probabilities, or the
        probability of *not* generating the offspring given maternal and
        candidate paternal genotypes

    Returns
    -------
    Array of likelihoods of paternity indexing sibships x candidates.
    """
    if not isinstance(offspring, genotypeArray):
        raise TypeError('offspring is not a genotypeArray')
    if not isinstance(mothers, genotypeArray):
        raise TypeError('mothers is not a genotypeArray')
    if not isinstance(males, genotypeArray):
        raise TypeError('males is not a genotypeArray')
    if len(partition) != offspring.size:
        raise ValueError('Length of partition does not match number of offspring.')
    if mu == 0:
        warn('Setting error rate to exactly zero causes clustering to be unstable.')

    # diploid genotypes
    offspring_diploid = offspring.geno.sum(2)
    maternal_diploid = mothers.geno.sum(2)
    male_diploid = males.geno.sum(2)

    # array of viable transition probabilities
    trans_prob_array = np.array([[[1,  0.5, 0  ],
                                  [0.5,0.25,0  ],
                                  [0,  0,   0  ]],
                                 [[0,  0.5, 1  ],
                                  [0.5,0.5, 0.5],
                                  [1,  0.5, 0  ]],
                                 [[0,  0,   0  ],
                                  [0,  0.25,0.5],
                                  [0,  0.5, 1  ]]])

    # array of probabilities for paternal genotypes when drawn from allele frequencies.
    af = np.array([allele_freqs**2,
              allele_freqs * (1-allele_freqs),
              (1-allele_freqs)**2])

    vals    = np.unique(partition)
    prob_s  = np.zeros([len(vals), males.size+1])
    missing = np.zeros(len(vals))

    for v in range(len(vals)):
        ix = np.where(partition == vals[v])[0]
        #print ix
        # empty arrays to stores probabilities.
        prob_f = np.zeros([3, males.size, offspring.nloci])
        prob_a = np.zeros([3, males.nloci])

        # positions of dropouts
        drop_m = (maternal_diploid[ix] < 0) + (offspring_diploid[ix] < 0) # offspring vs mothers

        geno =[0,1,2]
        for f in geno:
            prob_m = np.zeros(offspring_diploid[ix].shape)
            for m in geno:
                prob_o = np.zeros(offspring_diploid[ix].shape)
                for o in geno:
                    # the transition probability for the given genotypes.
                    if reciprocal: trans_prob = 1-trans_prob_array[o, m, f]
                    else:          trans_prob =   trans_prob_array[o, m, f]
                    # Probabilities that the observed offspring marker data match observed data.
                    pr_offs = np.zeros(offspring_diploid[ix].shape)
                    pr_offs[offspring_diploid[ix] == o] = 1-mu
                    pr_offs[offspring_diploid[ix] != o] = mu
                    prob_o = prob_o + (trans_prob * pr_offs * 1/3)
                # Probabilities that the observed maternal marker data match observed data.
                pr_mothers = np.zeros(maternal_diploid[ix].shape)
                pr_mothers[maternal_diploid[ix] == m] = 1-mu
                pr_mothers[maternal_diploid[ix] != m] = mu
                prob_m = prob_m + (prob_o * pr_mothers * 1/3)

            prob_m = prob_m / 3
            # Probabilities that the observed candidate male marker data match observed data.
            pr_males = np.zeros(male_diploid.shape)
            pr_males[male_diploid == f] = 1-mu
            pr_males[male_diploid != f] = mu

            prob_m[drop_m] = 1 # set loci with dropouts to prob=1. We will correct them these later.
            arr = np.log(prob_m).sum(0) + np.log(pr_males) # multiply by probability of paternal genotypes
            arr[male_diploid < 0] = 0 # allow for dropouts in the candidates
            # send to output.
            prob_f[f] = arr[np.newaxis]
            prob_a[f] = np.log(prob_m).sum(0) + np.log(af[f])

        # sum over possible paternal genotypes
        prob_f = alogsumexp(prob_f, axis=0)
        # correct for dropouts in father
        drop_f = drop_m[np.newaxis] + (male_diploid<0)[:,np.newaxis]
        corr = float(offspring.size*offspring.nloci) / (1-drop_f).sum(1).sum(1)
        prob_f = prob_f.sum(1) * corr
        # correct dropouts for missing fathers
        prob_a = alogsumexp(prob_a, axis=0).sum()
        corr = float(offspring.size*offspring.nloci) / (1-drop_m).sum()
        prob_a = prob_a * corr
        # send to output
        prob_s[v] = np.append(prob_f, prob_a)
    return(prob_s)



import numpy as np

def partition_accuracy(true, proposed, rtype='all'):
    """
    Returns the accuracy of a proposed partition with reference to a known true partition.
    
    Pr(A|B) is the probability of inferring relationship of type A between two individuals
    given that the true relationship is type B. This function estimates the total 
    proportion of pairwise relationships inferred correctly, as well as Pr(FS|FS), and 
    Pr(HS|HS). For half sibling arrays, Pr(FS|HS) = 1-Pr(HS|HS) and Pr(HS|FS) = 1-Pr(FS|FS).
    
    #ARGUMENTS
    true: true partition structure.
    
    proposed: the proposed partition structure.
    
    rtype: the measure of accuracy to be determine. This can be 'all' for overal accuracy,
    'fs' for Pr(FS|FS) or 'hs' for Pr{HS|HS).
    
    RETURNS
    A float between zero and one.
    """
    if rtype not in ['all', 'fs', 'hs']:
        print "rtype must be one of 'all', 'fs' or 'hs'."
        #return None

    real_FS = relation_matrix(true)
    part_FS = relation_matrix(proposed)
    r_given_r = real_FS == part_FS

    if rtype is 'all':
        correct   = (r_given_r * np.triu(np.ones(real_FS.shape), 1)).sum()
        total     = np.triu(np.ones(real_FS.shape), 1).sum()
        accuracy  = correct/total

    if rtype is 'fs':
        # Pr(FS|FS) depends on the correct relationships, conditioned on being a true FS relationship.
        fs_given_fs = r_given_r * real_FS
        correct     = (fs_given_fs * np.triu(real_FS,1)).sum()
        total       = np.triu(real_FS,1).sum()
        accuracy    = correct/total

    if rtype is 'hs':
        real_HS     = 1- real_FS # real halfsibling relationships
        hs_given_hs = r_given_r * real_HS
        correct     = (hs_given_hs * np.triu(real_HS, 1)).sum()
        total       = np.triu(real_HS,1).sum()
        accuracy    = correct/total

    return accuracy


import numpy as np

def pr_transition(offspring_diploid, maternal_diploid, male_diploid, offspring_genotype, maternal_genotype, male_genotype, mu):
    """
    Calculate the transition probability for a trio of putative diploid
    genotypes.

    Transition probabilities are then weight by the probability that the true
    offspring, mothers, and male genotypes match the input genotypes given
    observed marker data. Generally one would need to sum over all 27 possible
    combinations of genotypes.

    This function works with diploid genotypes, rather than a genotypeArray.
    Generally this is not called directly, but through lik_sampled_fathers() or 
    similar.

    Parameters
    ----------
    offspring_diploid, maternal_diploid, male_diploid: array
        arrays of diploid genotypes for the offspring, mothers and fathers.

    offspring_genotype, maternal_genotype, male_genotype: int
        0, 1 or 2 indicating homozygous, heterozygous or alternate homozygous
        genotype.

    mu: float
        point estimate of the genotyping error rate.

    Returns
    -------
    A 3-dimensional array of probabilities indexing offspring, candidate males,
    and loci. These are given in linear, rather than log space.
    """
    # an array of all possible transition probabilities indexed as [offspring, mother, father].
    trans_prob_array = np.array([[[1,  0.5, 0  ],
                                  [0.5,0.25,0  ],
                                  [0,  0,   0  ]],
                                 [[0,  0.5, 1  ],
                                  [0.5,0.5, 0.5],
                                  [1,  0.5, 0  ]],
                                 [[0,  0,   0  ],
                                  [0,  0.25,0.5],
                                  [0,  0.5, 1  ]]])
    # the transition probability for the given genotypes.
    trans_prob = trans_prob_array[offspring_genotype, maternal_genotype, male_genotype]

    # Probabilities that the observed offspring marker data match observed data.
    pr_offs = np.zeros([offspring_diploid.shape[0], offspring_diploid.shape[1]])
    pr_offs[offspring_diploid == offspring_genotype] = 1-mu
    pr_offs[offspring_diploid != offspring_genotype] = mu
    # Probabilities that the observed maternal marker data match observed data.
    pr_mothers = np.zeros([maternal_diploid.shape[0], maternal_diploid.shape[1]])
    pr_mothers[maternal_diploid == maternal_genotype] = 1-mu
    pr_mothers[maternal_diploid != maternal_genotype] = mu
    # Probabilities that the observed candidate male marker data match observed data.
    pr_males = np.zeros([male_diploid.shape[0], male_diploid.shape[1]])
    pr_males[male_diploid == male_genotype] = 1-mu
    pr_males[male_diploid != male_genotype] = mu

    return trans_prob * pr_males[np.newaxis] * pr_mothers[:,np.newaxis] * pr_offs[:,np.newaxis]



import numpy as np
from faps.paternityArray import paternityArray
from faps.sibshipCluster import sibshipCluster
from faps.do_clustering import do_clustering
from warnings import warn

def sibship_clustering(paternity_array, ndraws=1000, use_covariates = False, exp=False):
    """
    Cluster offspring into full sibship groups using hierarchical clustering.

    This first builds a dendrogram of relatedness between individuals, and pulls
    out every possible partition structure compatible with the dendrogram. The
    likelihood for each partition is also estimated by Monte Carlo simulations.

    This function is really just a wrapper for do_cluster() that allows for
    lists of paternityArray objects.

    If any paternityArray object contains only a single individual, clustering
    is meaningless, but an artifical sibshipCluster object is created with a
    single partition with log likelihood of 0.

    Parameters
    ----------
    paternity_array: paternityArray or dict
        Information on the likelihood of paternity for individuals in a maternal
        family. To analyse multiple maternal families in parallel,
        give a dictionary containing multiple named paternityArray objects.
    ndraws: int
        Number of Monte Carlo simulations to run for each partition.
    use_covariates: logical, optional
        If True, information on prbabilities associated with covariates stored
        in paternityArray objects are incorporated into sibship clustering.
    exp: logical, optional
        Indicates whether the probabilities of paternity should be exponentiated
        before calculating pairwise probabilities of sibships. This gives a
        speed boost if this is to be repeated many times in simulations, but
        there may be a cost to accuracy. Defaults to False for this reason.

    Returns
    -------
    A sibshipCluster object.

    Examples
    --------
    from faps import *
    import numpy as np

    # Generate a population of adults
    allele_freqs = np.random.uniform(0.3,0.5,50)
    adults = make_parents(20, allele_freqs)

    # Ecample with a single family
    # Mate the first adult to the next three.
    mother = adults.subset(0)
    progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
    # Create paternityArray
    patlik = paternity_array(progeny, mother, adults, mu=0.0013)
    # compare the effect of exponentiating likelihoods on the accuracy of sibship
    sc1 = sibship_clustering(patlik, exp=True)
    sc2 = sibship_clustering(patlik, exp=False)
    # see help(sibship_clustering) for explanation of the output
    sc1.accuracy(progeny, adults)
    sc2.accuracy(progeny, adults)

    # Example with multiple half-sib families
    progeny = make_offspring(parents = adults, dam_list=[7,7,1,8,8,0], sire_list=[2,4,6,3,0,7])
    # Split mothers and progeny up by half-sib array.
    mothers = adults.split(by=progeny.mothers, return_dict=True)
    progeny = progeny.split(by=progeny.mothers, return_dict=True)
    # Create paternity array for each family using dictionaries
    patlik = paternity_array(progeny, mothers, adults, mu=0.0013)
    # The dictionary is pass straight to sibship_clustering.
    sc = sibship_clustering(patlik)
    # Pull out info on individual families
    sc['base_7'].mlpartition
    # Since family 'base_0' contains only a single offspring, there is no
    # information about clustering. Compare with family `base_7`:
    sc['base_7'].linkage_matrix
    sc['base_0'].linkage_matrix
    """
    # If a single paternity array is given
    if isinstance(paternity_array, paternityArray):
        output = do_clustering(paternity_array,ndraws=ndraws, use_covariates= use_covariates, exp=exp)
    elif isinstance(paternity_array, list):
        if not all([isinstance(x, paternityArray) for x in paternity_array]):
            raise TypeError('Not all items in paternity_array are paternityArray objects.')
        output = [do_clustering(pa, ndraws=ndraws, use_covariates= use_covariates, exp=exp) for pa in paternity_array]
    # If dictionary is given
    elif(paternity_array, dict):
        # Check data types.
        if not all([isinstance(x, paternityArray) for x in paternity_array.values()]):
            raise TypeError('Not all items in paternity_array are paternityArray objects.')
        # Cluster each paternityArray.
        output = {}
        for pa in paternity_array.keys():
            if paternity_array[pa].offspring.shape[0] == 1:
                warn("Family {} has only one individual. A sibship array will be returned, but without a linkage matrix.".format(pa), stacklevel=2)
            sc = do_clustering(paternity_array = paternity_array[pa],
                               ndraws=ndraws,
                               use_covariates=use_covariates,
                               exp=exp)
            output[pa] = sc
    else:
        TypeError("paternity_array should be a paternityArray object, or list of paternityArray objects.")

    return output



import numpy as np
from faps.paternityArray import paternityArray
from faps.matingEvents import matingEvents
from faps.alogsumexp import alogsumexp
from faps.relation_matrix import relation_matrix
from faps.draw_fathers import draw_fathers
from faps.lik_partition import lik_partition
from pandas import DataFrame

class sibshipCluster(object):
    """
    Information on  the results of hierarchical clustering of an offspring array
    into full sibling groups.

    This is typcially not called directly, but through an instance of the function
    `paternity_array`.

    Parameters
    ----------
    paternity_array: paternityArray
        Object listing information on paternity of individuals.
    linkage_matrix: array
        Z-matrix from fastcluster.linkage.
    partitions: 2-d array
        Array of possible partition structures from the linkage matrix.
    lik_partitions: 1d-array
        Vector of log likelihoods for each partition structure.

    Returns
    -------
    prob_partitions: array
        log posterior probabilities of each partition structure (`lik_partitions`
        normalised to sum to one).
    mlpartition: list
        maximum-likelihood partition structure.
    noffspring: int
        Number of offspring in the array.
    npartitions: int
        Number of partitions recovered from the dendrogram.
    """
    def __init__(self, paternity_array, linkage_matrix, partitions, lik_partitions, paths, path_likelihoods, path_probs):
        self.candidates       = paternity_array.candidates
        self.paternity_array  = paternity_array.prob_array()
        self.partitions       = partitions
        self.linkage_matrix   = linkage_matrix
        self.lik_partitions   = lik_partitions
        self.prob_partitions  = self.lik_partitions - alogsumexp(self.lik_partitions)
        self.mlpartition      = self.partitions[np.where(self.lik_partitions == self.lik_partitions.max())[0][0]]
        self.noffspring       = len(self.mlpartition)
        self.npartitions      = len(self.lik_partitions)
        self.paths            = paths
        self.path_likelihoods = path_likelihoods
        self.path_probs       = path_probs

    def add_covariate(self, covariate):
        """
        Include a vector of (log) probabilities associated with covariates that
        provide additional information about paternity beyond that provided by
        genetic information (e.g. geographic distances).

        Parameters
        ----------
        covariate: 1-d array
            Vector of (log) probabilities of paternity based on non-genetic
            information, with one element for every candidate father. If this is a
            function of multiple sources they should be multiplied and included in
            this vector. If a list of offspring arrays have been supplied, this
            should be a list of vectors.

        Returns
        -------
        No output is printed; the covariate is added to the paternityArray as
        the attribute 'covariate'. Any existing information is overwritten. The
        vector is appended with an additional zero to allow for the final column
        of a the prob_array item in a paternityArray that accounts for the
        probability of missing fathers.
        """
        if isinstance(covariate, np.ndarray):
            if len(covariate.shape) > 1:
                raise ValueError("covariate should be a 1-d array, but has shape {}".format(covariate.shape))
            if len(self.candidates) != covariate.shape[0]:
                raise ValueError("Length of vector of covariates ({}) does not match the number of fathers ({})".format(len(self.candidates), covariate.shape[0]))
            if not all(covariate <= 0):
                warn("Not all values in covariate are less or equal to zero. Is it possible probabilities have not been log transformed?")
            covariate = np.append(covariate, 0)
            self.covariate = covariate
            return None
        else:
            raise TypeError("covariate should be a 1-d NumPy array.")

    def accuracy(self, progeny, adults):
        """
        Summarise statistics about the accuracy of sibship reconstruction when
        the true genealogy is known (for example from simulated families).

        Parameters
        ----------
        progeny: genotypeArray
            Genotype information on the progeny
        adults: genotypeArray
            Genotype information on the adults

        Returns
        -------
        Vector of statistics:
        0. Binary indiciator for whether the true partition was included in the
            sample of partitions.
        1. Difference in log likelihood for the maximum likelihood partition
            identified and the true partition. Positive values indicate that the
            ML partition had greater support.
        2. Posterior probability of the true number of families.
        3. Mean probabilities that a pair of full sibs are identified as full sibs.
        4. Mean probabilities that a pair of half sibs are identified as half sibs.
        5. Mean probabilities that a pair of half or full sibs are correctly
            assigned as such.
        6. Mean probability of paternity of the true sires for those sires who
            had been sampled (who had non-zero probability in the paternityArray).
        7. Mean probability that the sire had not been sampled for those
            individuals whose sire was truly absent (who had non-zero probability
            in the paternityArray).
        """
        # Was the true partition idenitifed by sibship clustering.
        true_part  = progeny.true_partition()
        nmatches   = np.array([(relation_matrix(self.partitions[x]) == relation_matrix(true_part)).sum()
                            for x in range(self.npartitions)])
        nmatches   = 1.0*nmatches / true_part.shape[0]**2 # divide by matrix size.
        true_found = int(1 in nmatches) # return 1 if the true partition is in self.partitions, otherwise zero

        delta_lik  = round(self.lik_partitions.max() - lik_partition(self.paternity_array, true_part),2) # delta lik
        # Prob correct number of families
        if len(self.nfamilies()) < progeny.nfamilies:
            nfamilies  = 0
        else:
            nfamilies = self.nfamilies()[progeny.nfamilies-1]
        # Pairwise sibship relationships
        full_sibs = self.partition_score(progeny.true_partition(), rtype='fs') # accuracy of full sibship reconstruction
        half_sibs = self.partition_score(progeny.true_partition(), rtype='hs') # accuracy of full sibship reconstruction
        all_sibs  = self.partition_score(progeny.true_partition(), rtype='all')# accuracy of full sibship reconstruction

        # Mean probability of paternity for true sires included in the sample.
        sire_ix = progeny.parent_index('f', adults.names) # positions of the true sires.
        dad_present = np.isfinite(self.paternity_array[range(progeny.size), sire_ix]) # index those sires with non-zero probability of paternity
        if any(dad_present):
            sire_probs = self.prob_paternity(sire_ix)
            sire_probs = sire_probs[dad_present]
            sire_probs = alogsumexp(sire_probs) - np.log(len(np.array(sire_probs))) # take mean
            sire_probs = sire_probs.squeeze()
        else:
            sire_probs = np.nan

        # Mean probability that the father is absent
        abs_probs = np.exp(self.prob_paternity(-1)).mean()

        output = np.array([true_found,
                           delta_lik,
                           np.round(nfamilies, 3),
                           np.round(full_sibs, 3),
                           np.round(all_sibs,  3),
                           np.round(half_sibs, 3),
                           np.round(np.exp(sire_probs),3),
                           np.round(abs_probs,3)])
        return output

    def nfamilies(self):
        """
        Posterior probability distribution of the number of full sibships in the
        array.

        Returns
        -------
        A vector of (exponentiated) probabilities that the array contains each
        integer value of full sibships from one to the maximum possible.
        """
        pprobs = np.exp(self.prob_partitions) # exponentiate partition likelihoods for simplicity
        # number of families in each partition
        nfams  = np.array([len(np.unique(i)) for i in self.partitions])
        # sum the probabilities of each partition containing each value of family number
        nprobs = np.array([pprobs[np.where(i == nfams)].sum() for i in range(1, self.noffspring+1)])
        nprobs = nprobs / nprobs.sum() # normalise
        return nprobs

    def family_size(self):
        """
        Multinomial posterior distribution of family sizes within the array,
        averaged over all partitions.

        Returns
        -------
        A vector of probabilities of observing a family of size *n*, where *n* is
        all integers from one to the number of offspring in the array.
        """
        pprobs = np.zeros(self.noffspring) # empty vector to store sizes
        # For each partition get the counts of each integer family size.
        for j in range(self.npartitions):
            counts = np.bincount(np.unique(self.partitions[j], return_counts=True)[1], minlength=self.noffspring+1).astype('float')[1:]
            counts = counts / counts.sum() # normalise to sum to one.
            pprobs+= counts * np.exp(self.prob_partitions[j]) # multiply by likelihood of the partition.
        return pprobs

    def mean_nfamilies(self):
        """
        Expected number of families given probabilities of each partition.
        """
        return np.average(np.arange(1, self.noffspring+1), weights=self.nfamilies())

    def full_sib_matrix(self, exp=False):
        """
        Create a matrix of log posterior probabilities that pairs of offspring
        are full siblings. This sums over the probabilities of each partition
        in which two individuals are full siblings, multiplied by the
        probability of that partition.

        By default, this creates a 3-dimensional matrix of log probabilities
        and sums using logsumexp, which preserves values in log space.
        Alternatively values can be exponentiated and summed directly, which
        will be less demanding on memory and the processor for large arrays,
        but probably at some cost to accuracy.

        Parameters
        ----------
        exp: logical
            If True, exponentiate log probabilities and sum these directly.
            Defaults to `False`.

        Returns
        -------
        An n*n array of log probabilities, where n is the number of offspring
        in the `sibshipCluster` object.
        """
        if exp is True:
            sibmat = np.zeros([self.noffspring, self.noffspring])
            for j in range(self.npartitions):
                sibmat+= np.exp(self.prob_partitions[j]) * np.array([self.partitions[j][i] == self.partitions[j] for i in range(self.noffspring)])

        if exp is False:
            sibmat = np.zeros([self.npartitions, self.noffspring, self.noffspring])
            with np.errstate(divide='ignore'):
                for j in range(self.npartitions):
                    sibmat[j] = self.prob_partitions[j] + np.log(np.array([self.partitions[j][i] == self.partitions[j] for i in range(self.noffspring)]))
            sibmat = alogsumexp(sibmat, 0)

        return sibmat

    def partition_score(self, reference, rtype='all'):
        """
        Returns the accuracy score for the `sibshipCluster` relative to a
        reference partition. This is usually the known true partition for a
        simulated array, where the partition is known.

        Accuracies can be calculated for only full-sibling pairs, only
        non-full-sibling pairs, or for all relationships.

        Parameters
        ----------
        reference: list
            Reference partition structure to refer to. This should be a list or
            vector of the same length as the number of offspring.
        rtype str
            Indicate whether to calculate accuracy for full-sibling, half-sibling,
            or all relationships. This is indicated by 'fs', 'hs' and 'all'
            respectively. Note that half-sibling really means 'not a full sibling'.
            The distinction is only meaningful for data sets with multiple
            half-sib families.

        Returns
        -------
        A float between zero and one.
        """

        if len(reference) != self.noffspring:
            raise ValueError("Reference partition should be the same length as the number of offspring.")
            return None

        obs = self.full_sib_matrix()
        rm  = relation_matrix(reference)

        # Matrix of ones and zeroes to reference elements for each relationship type.
        if   rtype is 'all': ix = np.triu(np.ones(rm.shape), 1)
        elif rtype is 'fs':  ix = np.triu(rm, 1)
        elif rtype is 'hs':  ix = np.triu(1-rm, 1)
        else:
            raise ValueError("rtype must be one of 'all', 'fs' or 'hs'.")
            return None

        # Get accuracy scores
        dev = abs(rm - np.exp(obs))
        dev = dev * ix
        dev = dev.sum() / ix.sum()

        return 1- dev

    def prob_paternity(self, reference=None):
        """
        Posterior probabilities of paternity for a set of reference fathers
        accounting for uncertainty in sibship structure.

        Parameters
        ----------
        reference: int, array-like, optional
            Indices for the candidates to return. If an integer, returns
            probabilties for a single candidate individual. To return
            probabilities for a vector of candidates, supply a list or array of
            integers of the same length as the number of offspring.

        Returns
        -------
        Array or vector of log posterior probabilities.
        """
        if reference is None:
            probs = np.zeros([self.npartitions, self.noffspring, self.paternity_array.shape[1]]) # empty matrix to store probs for each partitions
            for j in range(self.npartitions): # loop over partitions
                this_part = self.partitions[j]
                this_array = np.array([self.paternity_array[this_part[i] == this_part].sum(0) for i in range(self.noffspring)])
                this_array = this_array - alogsumexp(this_array,1)[:, np.newaxis] # normalise
                this_array+= self.prob_partitions[j] # multiply by probability of this partition.
                probs[j] = this_array
            probs = alogsumexp(probs, axis=0)
            return probs

        else:
            # If a vector of candidates has been supplied.
            if isinstance(reference, list) or isinstance(reference, np.ndarray):
                if len(reference) != self.noffspring:
                    raise ValueError("If the set of reference candidates is given as a list or numpy vector this must be of the same length as the number of offspring.")
                if any([reference[i] > self.paternity_array.shape[1] for i in range(len(reference))]):
                    raise ValueError("One or more indices in reference are greater than the number of candidates.")
            # If a single candidate has been supplied
            elif isinstance(reference, int):
                if reference > self.paternity_array.shape[1]:
                    raise ValueError("The index for the reference candidate is greater than the number of candidates.")
                else:
                    reference = [reference] * self.noffspring
            else:
                raise TypeError("reference should be given as a list or array of the same length as the number of offspring, or else a single integer.")

            probs = np.zeros([self.npartitions, self.noffspring]) # empty matrix to store probs for each partitions
            for j in range(self.npartitions): # loop over partitions
                this_part = self.partitions[j]
                this_array = np.array([self.paternity_array[this_part[i] == this_part].sum(0) for i in range(self.noffspring)])
                this_array = this_array - alogsumexp(this_array,1)[:, np.newaxis] # normalise
                probs[j]   = np.diag(this_array[:, reference]) # take only diagnical elements
            probs = probs + self.prob_partitions[:, np.newaxis]
            probs = alogsumexp(probs, 0)

            return probs

    def sires(self, labels='names'):
        """
        For every candidate drawn as a father in any partition, calculate the
        (log) probability that each sired at least one offspring.

        Probabilities are calculated as the sum of probabilties of each
        partition in which a candidate appears.

        Parameters
        ----------
        labels: array, optional
            Labels for candidates. By default, the names of candidates are used.
            If `labels=None`, the index positions in the list of candidates are
            used; this can be useful if you want to use indices to subset
            another dataset and link that to probabilities  of paternity.

        Returns
        -------
        DataFrame giving mother (taken from the keys of the input dictionary),
        fathers (inherited from each sibshipCluster object), probabilties of
        having sired at least one offspring, and the expected number of
        offspring.


        Examples
        --------
        from faps import *
        import numpy as np

        # Generate a population of adults
        allele_freqs = np.random.uniform(0.3,0.5,50)
        adults = make_parents(20, allele_freqs)

        # Ecample with a single family
        # Mate the first adult to the next three.
        mother = adults.subset(0)
        progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
        patlik = paternity_array(progeny, mother, adults, mu=0.0013)
        sc = sibship_clustering(patlik)

        sc.sires() # returns candidate names by default
        """
        if labels is 'names':
            labels = np.append(self.candidates, np.nan)

        # Get list of the unique set of fathers drawn for any partition.
        sires = np.unique([i for j in [x for y in self.paths.values() for x in y] for i in j])

        # Flatten lists of paths and their likelihoods
        flat_paths    = [x for y in self.paths.values() for x in y]
        flat_pathliks = [x for y in self.path_probs.values() for x in y if np.isfinite(x)]
        # For each sire drawn, find all paths he is involved in and sum their likelihoods.
        output = {}
        for s in sires:
            sx = [s in x for x in flat_paths]
            these_liks = np.array(flat_pathliks)[sx]
            output[s] = alogsumexp(these_liks)

        # Flip into data.frame
        output = DataFrame({
                    'position' : list(output.keys()),
                    'label'    : [labels[k] for k in output.keys()],
                    'log_prob' : list(output.values())
                })
        output['prob'] = np.exp(output['log_prob'])
        # Get the expected number of offspring for each mating event
        noffs = self.prob_paternity()[:, output['position']]
        noffs = alogsumexp(noffs, axis=0)
        output['offspring'] = np.exp(noffs)
        # Return DataFrame of information on sires
        return output

    def mating_events(self, paternity_array, unit_draws=1000, total_draws=10000, n_subsamples = 0, subsample_size = None, null_probs=None, use_covariates=False, return_names=True):
        """
        Sample plausible mating events from a sibshipCluster. These can then be
        used to infer mating patterns using and array of phenotypes for each male.

        Indices for possible fathers are drawn based on genetic data, other pertinent
        information (such as distance between parents), or a combination of these
        types. Mating events are drawn for each partition in a `sibshipCluster` object.
        These samples are then resampled in proportion to a weight on each unit.
        To assess uncertainty, the aggregate mating events can be subsampled.

        You can also supply an array of probabilities that a given father is the sire
        based on some appropriare null model, which can then be compared with observed
        mating patterns. For example, to test whether observed patterns could be
        explained by random mating but with spatial clustering, you could supply an
        array of probabilities derived from an appropriate dispersal function.

        Parameters
        ----------
        paternity_array: paternityArray
            Object listing information on paternity of individuals. This should be the
            same paternityArray object used to construct the sibshipCluster objects,
            or else results will be meaningless.
        unit_draws: int
            Number of mating events to sample for each partition.
        total_draws: int
            Total number of mating events to resample for each partition.
        n_subsamples: int, optional
            Number of subsamples to draw from the total mating events. Defaults
            to zero.
        subsample_size: int, optional
            Number of mating events in each subsample. Defaults to
            0.1*total_draws.
        null_probs: array, optional
            1-d Array of probabilities for paternity if this were not based on
            marker data. This should have an element for every candidate father.
        use_covariates: logical, optional
            If True, information on prbabilities associated with covariates
            stored in paternityArray objects are incorporated into weights for
            drawing likely fathers.
        return_names: logical, optional
            If True, unit_events and total_events are returned as the names of
            the candidates drawn for each partition. If False, their integer
            positions in the list of candidates is returned.

        Returns
        -------
        A matingEvents object.

        Examples
        --------
        import numpy as np
        from faps import *

        # simulate genotype data
        allele_freqs = np.random.uniform(0.1, 0.5, 50)
        males = make_parents(250, allele_freqs)
        offspring = make_sibships(males, 0, range(1,5), 5)
        mothers = males.subset(offspring.parent_index('m', males.names))

        # Creat paternityArrays and cluster into full sibships.
        patlik = paternity_array(offspring, mothers, males, mu=0.0015)
        patlik.add_covariate(np.log(np.repeat(1.0/males.size, males.size)))
        sc = sibship_clustering(patlik)

        # Basic usage
        sc.mating_events(patlik)

        # Including probabilities from covariates stored in the paternityArray.
        sc.mating_events(patlik, use_covariates=True)

        # Generate a sample in proportion to some other probability function not including genetic information
        null_probs = np.repeat(1.0/males.size, males.size)
        sc.mating_events(patlik, null_probs=null_probs)
        """
        if not isinstance(paternity_array, paternityArray):
            raise TypeError('paternity_array should be a paternityArray.')
        elif paternity_array.prob_array().shape != self.paternity_array.shape:
            raise ValueError('Number of offspring in paternity ({}) does not match the sibshipCluster ({}).'.format(paternity_array.prob_array().shape, self.paternity_array.shape))

        if not isinstance(unit_draws, int):
            raise ValueError('unit_draws should be an integer.')
        if not isinstance(total_draws, int):
            raise ValueError('total_draws should be an integer.')

        # only consider partitions that would account for at least one mating event.
        valid_ix = np.around(total_draws * np.exp(self.prob_partitions)) >= 1
        valid_partitions = self.partitions[valid_ix]
        # Get names for those
        unit_names =  ['partition_' + str(x) for x in np.where(valid_ix)[0]]

        # mating_events draws a sample of fathers from either genetic data,
        # or else from some array of probabilties based on non-genetic data.
        # Draw fathers from genetic data:
        if null_probs is None:
            # draw mating events for each partition.
            unit_events = {}
            for k,v in zip(unit_names, valid_partitions):
                d = draw_fathers(v, paternity_array=paternity_array, ndraws=unit_draws, use_covariates=use_covariates)
                unit_events[k] = d
        # Draw fathers from non-genetic data.
        elif isinstance(null_probs, np.ndarray):
            if len(null_probs.shape) != 1:
                raise ValueError("null_probs should be a 1-d vector, but has shape {}".format("null_probs.shape"))
            elif len(null_probs) != len(paternity_array.candidates):
                raise ValueError("null_probs should have an element for every candidate father, but has length {}.".format(len(null_probs)))
            else:
                unit_events = {}
                for k,v in zip(unit_names, valid_partitions):
                    unit_events[k] = draw_fathers(v, null_probs=null_probs, ndraws=unit_draws)
        else:
            raise TypeError('null_probs is of type {}. If supplied, this should is supplied it should be an array.'.format(type(null_probs)))
        # Resample mating events for each partition weighted by the probability
        # for that partition.
        # First, get an set of integer number of draws for each partition.
        unit_weights = np.around(np.exp(self.prob_partitions[valid_ix]) * total_draws).astype('int')
        unit_weights = {k:v for k,v in zip(unit_names, unit_weights)}
        # Resample unit_events proportional to the prob of each unit.
        total_events = [np.random.choice(a=v, size=unit_weights[k], replace=True) for k,v in unit_events.items()]
        total_events = [item for sublist in total_events for item in sublist]
        total_events = np.array(total_events)

        # Subsample mating events.
        if isinstance(n_subsamples, int):
            if subsample_size is None:
                subsample_size = np.around(total_draws*0.1).astype('int')
            elif not isinstance(subsample_size, int):
                raise TypeError("subsample_size should be an integer or None.")
            if subsample_size >= total_draws:
                raise ValueError('Number of subsamples must be smaller than total_draws')
            sub_events = np.random.choice(total_events, n_subsamples*subsample_size, replace=True)
            sub_events = sub_events.reshape([n_subsamples, subsample_size])
        else:
            raise ValueError('n_subsamples should be an integer.')


        return matingEvents(unit_names=unit_names,
                            candidates=paternity_array.candidates,
                            unit_weights=unit_weights,
                            unit_events=unit_events,
                            total_events=total_events,
                            subsamples=sub_events)



from time import time, localtime, asctime
import numpy as np
from faps.make_parents import make_parents
from faps.make_sibships import make_sibships
from faps.paternityArray import paternityArray
from faps.paternity_array import paternity_array
from faps.sibshipCluster import sibshipCluster
from faps.sibship_clustering import sibship_clustering
from faps.lik_partition import lik_partition
from faps.relation_matrix import relation_matrix

def make_generation(allele_freqs, candidates, sires, offspring, missing_loci=0, mu_real=0, mu_input=0, unsampled_real=None, unsampled_input=0, selfing_rate=0, cluster_draws=1000, exp_clusters=True, return_paternities=False, return_clusters=False, counter='NA'):
    """
    Accuracy of sibship and paternity assignment for a single simulated half-sibling array.

    Simulates a population of adults and generates a half-sibling array. Returns information
    on whether the true partition was among those identified by clustering, accuracy of
    sibling relationships and posterior probabilities of paternity.

    Parameters
    ----------
    allele_freqs: list
        Vector of allele frequencies.
    candidates: int
        Adult population size.
    sires: int or list
        Number of males to sire full sibships, or a list of males. The mother
        is always the first candidate in the array, and is indexed as position zero.
    offspring: int or list
        Full sibship sizes. If a single integer has been passed to `sires`,
        this must be an integer indicating full sibship size, which is constant for
        all full sibships. If a list is passed to `sires`, this should be a list of
        the same length indicating the size of each individual full-sib family.
    missing_loci: float
        Real diploid allelic dropout rate between 0 and 1.
    mu_real: float
        Real haploid genotyping error rate between 0 and 1.
    mu_input: float
        The genotyping error rate used to construct the `paternityArray`. Values must be
        between 0 and 1.
    unsampled_real: float or list
        Individuals to remove from the list of candidate males. If a float between
        zero and one is given, this indicates a proportion of candidates which will
        be removed at random, reflecting incomplete sampling. If a list of integers is
        given, this indicates specific indices of candidates to be removed, such as
        a known true sire, or the mother.
    unsampled_input: float or list
        Value for the proportion of unsampled fathers to be used in sibship clustering.
        Values must be between 0 and 1.
    selfing_rate: float
        Value for the selfing rate to be used in sibship clustering. Values must be
        between 0 and 1.
    cluster_draws: int, optional
        Number of Monte Carlo draws in sibship clustering.
    exp_clusters: logical, optional
        If `True` exponentiate pairwise sibship matrix to speed up sibship clustering
        with a potential cost to accuracy. See `sibship_clustering` for details.
    return_paternities: logical, optional
        If `True` the array of paternity probabilities object is returned for each
        simulated array.
    return_clusters: bool, optional
        If `True`, a `sibshipCluster` object is returned for each simulated array.
    verbose: bool, optional
        If true prints output on parameter input and simulation progress.

    Returns
    -------
    An array listing:
        0. simulation index (when called from `make_power`)
        1. number of loci
        2. allele frequencies
        3. number of candidates
        4. number of sires
        5. number of offspring
        6. genotype dropout rate
        7. real genotype error rate
        8. input genotype error rate
        9. proportion of missing candidates
        10. input value for the presumed number of missing fathers
        11. input selfing rate
        12. time in seconds to create paternityArray
        13. time in seconds to perform clustering
        14. binary indiciator for whether the true partition was included in the
            sample of partitions.
        15. difference in log likelihood for the maximum likelihood partition
            identified and the true partition. Positive values indicate that the
            ML partition had greater support.
        16. posterior probability of the true number of families.
        17. mean probabilities that a pair of full sibs are identified as full sibs.
        18. mean probabilities that a pair of half sibs are identified as half sibs.
        19. mean probabilities that a pair of half or full sibs are correctly
            assigned as such.
        20. mean probability of paternity of the true sires for those sires who
            had been sampled (who had non-zero probability in the paternityArray).
        21. mean probability that the sire had not been sampled for those
            individuals whose sire was truly absent (who had non-zero probability
            in the paternityArray).

    If return_clusters is true, a list is returned containing the array described above and the sibshipCluster.

    Examples
    --------
    import numpy as np

    allele_freqs = np.random.uniform(0.3,0.5,80)
    make_generation(allele_freqs, 100, [0,2,3], [5,4,3], mu_real=0.0015, mu_input = 0.0015, unsampled_real=0.05, selfing_rate = 0.3)
    """
    # simulate population
    adults = make_parents(candidates, allele_freqs) # parents
    # Generate the progeny
    if isinstance(sires, int):
        progeny = make_sibships(
        parents = adults,
        dam = 0,
        sires = np.arange(1,sires+1).tolist(),
        family_size = offspring
    )
    elif isinstance(sires, list) or isinstance(sires, np.ndarray):
        progeny = make_sibships(
            progeny = adults,
            dam = 0,
            sires = sires,
            family_size = offspring
        )
    # introduce genotyping errors
    adults = adults.mutations(mu_real).dropouts(missing_loci)
    progeny= progeny.mutations(mu_real).dropouts(missing_loci)
    # reference mothers
    mothers = adults.subset(progeny.parent_index('m', adults.names))

    # Paternity array
    t1      = time()
    patlik  = paternity_array(
        progeny,
        mothers,
        adults,
        mu = mu_input,
        purge = unsampled_real,
        selfing_rate = selfing_rate,
        missing_parents = unsampled_input
    )
    #patlik.prob_array = patlik.adjust_prob_array(unsampled_real, unsampled_input, selfing_rate)
    #Sibship clustering
    t2      = time()
    sc      = sibship_clustering(
                  paternity_array = patlik,
                  ndraws = cluster_draws,
                  exp=exp_clusters,
                  use_covariates=False)
    t3      = time()
    # Determine whether the true partition is among those inferred by clustering.
    # For each partition count up how many pairwise relationships were inferred correct.
    true_part  = progeny.true_partition()
    nmatches   = np.array([(relation_matrix(sc.partitions[x]) == relation_matrix(true_part)).sum()
                           for x in range(sc.npartitions)])
    nmatches   = 1.0*nmatches / true_part.shape[0]**2 # divide by matrix size.
    true_found = int(1 in nmatches) # return 1 if the true partition is in sc.partitions, otherwise zero

    # value to return for the proportion of missing fathers
    if   isinstance(unsampled_real, float): sprop = unsampled_real
    elif isinstance(unsampled_real, list):  sprop = float(len(unsampled_real)) / candidates
    else: sprop = 0

    # a list of simulation data to return.
    output = np.array([counter,
                len(adults.allele_freqs()), # number of loci
                round(adults.allele_freqs().mean(),3), #allele_freqs
                candidates, # number of candidates
                len(np.unique(progeny.fathers)), # sires
                progeny.size, # offspring
                missing_loci,
                mu_real,
                mu_input,
                sprop, # proportion missing fathers
                unsampled_input,
                selfing_rate,
                # timing
                round(t2-t1,3), # time in seconds to create paternityArray
                round(t3-t2,3)  # time in seconds to perform clustering
                ])
    output = np.append(output, sc.accuracy(progeny, adults))

    outlist = [output]
    if return_paternities: outlist = outlist + [patlik]
    if return_clusters:    outlist = outlist + [sc]
    return outlist



from warnings import warn

def sibship_transitions(families, paternity_array, males, exp=False):
    """
    Transition probabilities (really log likelihoods) that each male is the sire
    of a set of putative full sibships.

    This differs from transition_probabilities in that sibship_transitions
    multiplies over sibships *before* summing over paternal genotypes. This
    ought to be more accurate because any genotyping error in the candidate 
    fathers is not multiplied over offspring, but incurs a loss in speed of
    about one order of magnitude.

    Parameters
    ----------
    families: list of vectors.
        A list of vectors, each giving the indices of one or more offspring in
        a putative full sibship. This is usually based on a set of partitions
        generated using sibship_partitions.
    paternity_array: paternityArray
        paternityArray object used to generate families.
    males: genotypeArray
        genotypeArray object used to generate families.
    exp: boolean, optional
        If True, likelihood arrays are exponentiated. This confers a substantial
        speed boost, but is unstable if the sample includes families with very
        low likelihood, because it is necessary to multiply over many small
        numbers, causing underflow.

    Returns
    -------
    An array of likelihoods that each full sibship was sired by each candidate
    male, indexed as n.families x n/candidates+1. The extra column is for the 
    likelihood that paternal alleles are drawn from population allele frequencies.
    Output is not normalised over candidates.

    Example
    -------
    import numpy as np
    from faps import *

    # generate 4 families of 5 offspring
    allele_freqs = np.random.uniform(0.3, 0.5, 50)
    males = make_parents(200, allele_freqs)
    offspring = make_sibships(males, 0, range(1,5), 5)

    # Add muations
    mu = 0.0013
    males = males.dropouts(0.01).mutations(mu)
    offspring= offspring.dropouts(0.01).mutations(mu)

    mothers = males.subset(offspring.parent_index('m', males.names))
    patlik = paternity_array(offspring, mothers, males, allele_freqs, mu)
    # Generate possible sets of families from fullpairs
    partition_sample = sibship_partitions(patlik)

    #get a list of all unique families.
    unique_families = [np.where(p == val)[0] for p in partition_sample for val in np.unique(p)]
    unique_families = set([tuple(up) for up in unique_families]) # coerce to tuple to create set
    unique_families = [np.array(up) for up in unique_families] # coerce back to numpy arrays

    sibship_transitions(unique_families, patlik, males)

    """
    # check types.
    if not isinstance(males, genotypeArray):
        TypeError("Argument paternity_array should be the paternityArray used to generate the list of plausible families.")
    if not isinstance(paternity_array, paternityArray):
        TypeError("Argument paternity_array should be the paternityArray used to generate the list of plausible families.")
    if not all(paternity_array.candidates == males.names):
        warn("Not all names of candidate fathers in the paternity array match those in the genotype array.\nEnsure both arrays are really those used to generate the list of families.")
    if paternity_array.by_locus is None:
        ValueError("paternity_array does not include ")

    if isinstance(patlik.by_locus, np.ndarray):
        if patlik.by_locus.shape == (3, len(patlik.offspring), males.nloci):
            ValueError("The attribute by_locus in the genotype array does not match the genotype array.")
    else:
        TypeError("The attribute by_locus in the genotype array does not exist, or is not a NumPy array.")
 
    # set mutational error rate.
    mu = paternity_array.mu
    if mu == 0:
        mu = 10**-12
        warn('Setting error rate to exactly zero causes clustering to be unstable. mu set to 10e-12')
    
    male_diploid = males.geno.sum(2)
    # create matrix to correct for dropouts.
    drop_f = (male_diploid < 0)
    corr   = 1/(1 - drop_f.mean(1))[np.newaxis]
    
    # likelihood for each unique family given maternal and offspring genotypes. 
    prob_u = np.array([paternity_array.by_locus[:,up].sum(1) for up in families])
    
        # array of possible allele frequency probabilities
    af = np.array([allele_freqs ** 2,
                   allele_freqs * (1-allele_freqs),
                   (1-allele_freqs)**2])
    af = np.log(af)
    prob_a = alogsumexp(prob_u + af[np.newaxis], 1).sum(1)
    # get the likelihood for each possible paternal genotype, exponentiating likelihoods.
    if exp:
        prob_f = np.zeros([len(families), males.size, males.nloci])
        prob_u = np.exp(prob_u)
        for f in [0,1,2]:
            pr_males = np.zeros(male_diploid.shape)
            pr_males[male_diploid == f] = 1-mu
            pr_males[male_diploid != f] = mu
            pr_males[drop_f] = 1
            prob_f += prob_u[:,f][:, np.newaxis] * pr_males[np.newaxis]
        with np.errstate(divide='ignore'):
            siblik = np.log(prob_f.prod(2))
    
    # get the likelihood for each possible paternal genotype, without exponentiating likelihoods.
    else:
        prob_f = np.zeros([3, len(families), males.size, males.nloci])
        for f in [0,1,2]:
            pr_males = np.zeros(male_diploid.shape)
            pr_males[male_diploid == f] = 1-mu
            pr_males[male_diploid != f] = mu
            pr_males[drop_f] = 1
            prob_f[f] = prob_u[:,f][:, np.newaxis] + np.log(pr_males)[np.newaxis]
        siblik = alogsumexp(prob_f, axis=0).sum(2)
    
    siblik = siblik + np.log(corr)
    
    return np.column_stack([siblik, prob_a])


import numpy as np

def alogsumexp(logarray, axis=0):
    """
    Calculate the sum of an array of values which are in log space. If an axis is
    specified, the sum across this axis is given.
    
    Parameters
    ----------
    logarray: array
        An array of log values to be summed.
    axis: int, optional
        Axis to sum across.
    
    Returns
    -------
    Either a float corresponding to the log of summed values, or the vector of
    sums across a specific axis.
    """
    with np.errstate(invalid='ignore', divide='ignore'): # turn off warnings about dividing by zeros.
        shp       = list(logarray.shape)
        shp[axis] = 1
        maxvals   = logarray.max(axis=axis)
        sumarray  = np.log(np.sum(np.exp(logarray - maxvals.reshape(shp)), axis = axis)) + maxvals
        if not isinstance(sumarray, float): 
            sumarray[np.isnan(sumarray)] = np.log(0)
        return sumarray



import numpy as np
from warnings import warn
from faps.calculate_geno_probs import calculate_geno_probs

class genotypeArray(object):
    """
    Genotype information about a sample of individuals, the identity of any known
    relatives, and metadata about the dataset.

    Currently only SNP data are supported. Data are recorded as integers: zeros
    and twos indicate opposing homozygous genotypes, and ones heterozygous
    genotypes.

    Parameters
    ----------
    geno: array
        3-dimensional array of genotype data indexing (1) individual, (2) locus,
        and (3) chromosome pair.
    names: array-like
        Unique identifiers for each individual.
    mothers: array-like
        Identifiers of the mother of each individual, if known.
    fathers: array-like
        Identifiers of the father of each individual, if known.
    markers: array-like, optional
        Marker names.

    Returns
    -------
    size: int
        Number of indivudals
    nloci: int
        Number of markers in the dataset
    parents: array
        Names of parent pairs for each indivudals.
    families: array
        List of unique full-sib families, based on the names of parents.
    nfamilies: int
        Number of full-sib families, based on the names of parents.
    """
    def __init__(self, geno, geno_probs, names, mothers, fathers, markers=None):
        self.geno      = geno
        self.geno_probs= geno_probs
        self.names     = names
        self.mothers   = mothers
        self.fathers   = fathers
        self.markers   = markers
        self.size      = self.geno.shape[0]
        self.nloci     = self.geno.shape[1]
        self.parents   = np.array([str(self.mothers[o]) + '/' + str(self.fathers[o]) for o in range(self.size)])
        self.families  = np.unique(self.parents)
        self.nfamilies = len(self.families)

    def allele_freqs(self):
        """
        Allele frequencies of each locus. The reference allele is whichever is
        labelled as 1.
        """
        diploid = self.geno.sum(2) * 0.5
        return np.nanmean(diploid, axis = 0)

    def drop(self, individuals):
        """
        Remove specific individuals from the genotype array.

        Parameters
        ----------
        individuals: an integer or list of integers indexing the individuals to be removed.

        Returns
        -------
        A genotype array with the target individuals removed.
        """
        # If indices are given
        if isinstance(individuals, int):
            if individuals > self.size or individuals < 0:
                raise ValueError("The index for the individual to drop is greater than the number of individuals, or less than zero.")
            individuals = [individuals]
        # If a names are given, find the positions in the list of names
        if isinstance(individuals, str):
            if individuals not in self.names:
                raise ValueError("The name for the individual to drop is not present in the list of names for this genotypeArray object.")
            individuals = [individuals]
        if all(isinstance(x, str) for x in individuals):
            if (~np.isin(individuals, self.names)).any():
                raise ValueError("The name of one or more individuals to drop is not present in the list of names for this genotypeArray object.")
            individuals = [np.where(self.names == x)[0][0] for x in individuals]

        if any([i < 0 or i > self.size for i in individuals]):
            raise ValueError("One or more indices for individuals to drop is greater than the total number of individuals, or less than zero.")

        # Indices of candidates to keep.
        new_index = ~np.isin(np.arange(self.size), individuals)
        new_index = np.where(new_index)[0]

        # create new genotypeArray.
        output = genotypeArray(
            geno = self.geno[new_index],
            geno_probs = self.geno_probs[new_index],
            names = self.names[new_index],
            mothers = self.mothers[new_index],
            fathers = self.fathers[new_index]
            )
        return output

    def dropouts(self, dr):
        """
        Add allelic dropouts to an array of genotypic data.

        Parameters
        ----------
        dr: float
            Diploid dropout rate.

        Returns
        -------
        A copy of the input genotype data, but with dropouts (shown as nan).
        """

        # pick data points to dropout
        vals = np.random.binomial(1, dr, self.size * self.nloci)
        positions = np.reshape(vals, [ self.size, self.nloci])
        positions = positions.astype("bool")

        # make a copy of the genotype data, just in case
        new_geno       = np.copy(self.geno).astype(float)
        new_geno_probs = np.copy(self.geno_probs).astype(float)
        # insert missing data into parental genotypes
        new_geno[positions] = np.nan
        new_geno_probs[positions] = np.nan

        output = genotypeArray(
            geno = new_geno,
            geno_probs = new_geno_probs,
            names = self.names,
            mothers = self.mothers,
            fathers = self.fathers
            )

        return output

    def heterozygosity(self, by='marker'):
        """
        Mean heterozygosity, either averaged across markers for each individual,
        or across individuals for each marker.

        Parameters
        ----------
        by: str
            If 'individual', values are returned averaged across markers for each
            individual. If 'marker' values are returned averaged across
            individuals for each marker. Defaults to 'marker'.

        Returns
        -------
        Vector of floats.
        """
        if by == 'marker' or by == 0:
            return (self.geno.sum(2) == 1).mean(0)
        elif by == 'individual' or by == 1:
            return (self.geno.sum(2) == 1).mean(1)
        else:
            raise ValueError("`by` should be either 'marker' or 'individual'.")

    def missing_data(self, by='marker'):
        """
        Mean genotype drop-out rate, either averaged across markers for each individual,
        or across individuals for each marker.

        Parameters
        ----------
        by: str
            If 'individual', values are returned averaged across markers for each
            individual. If 'marker' values are returned averaged across
            individuals for each marker. Defaults to 'marker'.

        Returns
        -------
        Vector of floats.
        """
        d = np.copy(self.geno).astype(float)
        d[d == -9] = np.nan
        if by == 'marker' or by == 0:
            return np.isnan(d[:,:,0]).mean(0)
        elif by == 'individual' or by == 1:
            return np.isnan(d[:,:,0]).mean(1)
        else:
            raise ValueError("`by` should be either 'marker' or 'individual'.")

    def mutations(self, mu):
        """
        Introduce mutations at random to an array of genotype data for multiple individuals.
        For all alleles present draw mutations given error rate mu, then swap zeroes and
        ones in the array.

        Parameters
        ----------
        mu: float
            Haploid genotyping error rate.

        Returns
        -------
        A copy of the input genotype data, but with point mutations added.
        """
        # make a copy of the data, and make it an integer
        new_alleles = np.copy(self.geno)

        # for an array of the same shape as newAlleles, draw mutations at each
        # position with probability mu.
        vals = np.random.binomial(1, mu, self.size * self.nloci * 2)
        mutate = np.reshape(vals, [ self.size, self.nloci, 2])
        mutate = (mutate == 1)
        # swap zeroes and ones.
        new_alleles[mutate] ^= 1

        # Apply to geno_probs
        new_geno_probs = calculate_geno_probs(new_alleles, mu=mu)

        output = genotypeArray(
            geno = new_alleles,
            geno_probs = new_geno_probs,
            names = self.names,
            mothers= self.mothers,
            fathers = self.fathers
            )

        return output

    def parent_index(self, parent, parent_names):
        """
        Finds the position of parental names in a vector of possible parental
        names. This can be the name of the mother or the father.

        This is essentially a convenient wrapper for np.where().

        Parameters
        ----------
        parent: str
            A string indicating whether the offspring's mother or father is to
            be located. Valid arguments are 'mother', 'father' and 'parents',  or
            equivalently 'm' and 'f' respectively.

        parent_names: array
            1-d array of parental names to be found in the lists of mothers,
            father or parents.

        Returns
        -------
        A list of positions of the parent for each entry in offs_names.

        Example
        -------
        from faps import *
        import numpy as np
        # create genotypes
        allele_freqs = np.random.uniform(0.3,0.5,10)
        parents = make_parents(5, allele_freqs, family_name='my_population')
        progeny = make_sibships(mypop, 0, [1,2,3], 4, 'myprogeny')

        progeny.parent_index('mother', parents.names) # position of the mother
        progeny.parent_index('father', parents.names) # positions of the fathers
        """
        if parent is 'mother' or parent is 'm':
            return [np.where(parent_names == x)[0][0] for x in self.mothers]
        if parent is 'father' or parent is 'f':
            return [np.where(parent_names == x)[0][0] for x in self.fathers]
        if parent is 'parents' or parent is 'p':
            return [np.where(parent_names == x)[0][0] for x in self.parents]
        else:
            raise ValueError("parent must be 'mother', 'father', or 'parents'.")

    def split(self, by, return_dict=True):
        """
        Split up a gentotypeArray into groups according to some grouping
        factor. For example, divide an array containing genotype data for
        multiple half-sibling arrays by the ID of their mothers.

        Parameters
        ----------
        by: array-like
            Vector containing grouping labels for each individual
        return_dict: logical
            If True, the output is returned as a dictionary of genotypeArray
            objects indexed by entries in `by`. If False, a list is returned.
            Defaults to True.

        Returns
        -------
        A dictionary of genotypeArray objects.

        Examples
        --------
        from faps import *
        import numpy as np

        # Generate a population of adults
        allele_freqs = np.random.uniform(0.3,0.5,50)
        adults = make_parents(20, allele_freqs)

        # Mate the first adult to the next three.
        mother = adults.subset(0)
        progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')

        # Split by fathers
        progeny.split(progeny.fathers)
        """
        groups = np.unique(by)
        ix = [np.where(by == groups[i])[0] for i in range(len(groups))]
        if return_dict:
            output = {k:self.subset(i) for k,i in zip(groups, ix)}
        else:
            output = [self.subset(i) for i in ix]
        return output

    def subset(self, individuals=None, loci=None):
        """
        Subset the genotype array by individual or number of loci.

        To subset by both individuals and loci, call the function twice.

        Parameters
        ----------
        individuals: int, str, or a vector thereof
            Either a list of individual names, or integers indexing those
            individuals.

        loci: int, str, or a vector thereof
            Either a list of individual markers, or integers indexing those
            individuals.
        Returns
        -------
        A genotype array with only the target individuals included.
        """
        # if no subsetting indices are given, return the whole object.
        if individuals is None and loci is None:
            return self

        # Subset individuals if necessary
        if individuals is not None:
            # If only a single individual is given, make it a list.
            if isinstance(individuals, int):
                individuals = [individuals]
            if isinstance(individuals, str):
                individuals = [individuals]
            # If a names are given, find the positions in the list of names
            if all(isinstance(x, str) for x in individuals):
                individuals = [np.where(self.names == x)[0][0] for x in individuals]
            # Subset the genotypeArray
            output = genotypeArray(
                geno       = self.geno[individuals],
                geno_probs = self.geno_probs[individuals],
                names      = self.names[individuals],
                mothers    = self.mothers[individuals],
                fathers    = self.fathers[individuals],
                markers    = self.markers)

        # Subset loci if necessary
        if loci is not None:
            # If only a single locus is given, make it a list.
            if isinstance(loci, int):
                loci = [loci]
            if isinstance(loci, str):
                loci = [loci]
            # If a marker names are given, find the positions in the list of names
            if all(isinstance(x, str) for x in loci):
                loci = [np.where(self.markersz == x)[0][0] for x in loci]
            # If an array of boolean values are given, make to a list
            if isinstance(loci, np.ndarray):
                if np.result_type(loci) == 'bool':
                    loci = np.arange(len(loci))[loci]
                loci = loci.tolist()
            # Subset the genotypeArray
            output = genotypeArray(
                geno       = self.geno[:,loci],
                geno_probs = self.geno_probs[:,loci],
                names      = self.names,
                mothers    = self.mothers,
                fathers    = self.fathers,
                markers    = self.markers[loci])
        return output

    def true_partition(self):
        """
        For families of known parentage, usually simulated data, create a full sibship partition
        vector from the identities of known mothers and fathers contained in variables 'mothers'
        and 'fathers' of a genotype array.

        If one or more individuals has at least one missing parent they will be assigned to the
        same full sibship group.

        Returns
        -------
        An array of integers with an entry for each offspring individual. Individuals are labelled
        according to their full sibling group.
        """
        if 'NA' in self.mothers or 'NA' in self.fathers:
            warn('Warning: one or more individuals has at least one parent of unkown identity.')
            warn('All such individuals will be assigned to the same sibship group.')

        # concatenate mother and father names to create a vector of parent pairs.
        #parentage = np.array([str(self.mothers[o]) + '/' + str(self.fathers[o]) for o in range(noffs)])
        possible_families = np.unique(self.parents) # get a list of all unique parent pairs

        partitions = np.zeros(self.size).astype('int') # empty vector of zeros.
        for o in range(self.nfamilies):
            # for every possible family label individuals with an identical integer.
            partitions[self.parents == possible_families[o]] += o

        return partitions

    def write(self, filename, delimiter = ','):
	    """
	    Write data in a genotypeArray to disk. Columns for known mothers and fathers
	    are included, even if these are all NA.

	    Parameters
	    ----------
	    filename: stre
	    	System path to write to.
	    delimiter: str, optional
	    	Column delimiter. Defaults to commas to generate CSV files.

	    Returns
	    -------
	    A text file at the specified path.

	    """
	    # Names of individuals, plus mothers and fathers.
	    nms = np.column_stack([self.names, self.mothers, self.fathers])
	    # format genotype data as a strings
	    output = self.geno.sum(2).astype('str')
	    output[output == '-18'] = 'NA' # coerce missing data to NA

	    output = np.concatenate([nms, output], axis=1)
	    header = 'ID,mother,father,' + ','.join(self.markers)
	    np.savetxt(filename, output, delimiter=delimiter, fmt="%s", header=header, comments='')



import numpy as np
from faps.alogsumexp import alogsumexp
from faps.squash_siblings import squash_siblings
from faps.unique_rows import unique_rows

def lik_partition(paternity_probs, partitions, includes_null=True, ndraws=1000):
    """
    Find the likelihood of a partition by sampling fathers for each sibship proportional to his probability.
    This then removes duplicate sets of fathers, and sets for which more than one family shares a father.

    Parameters
    ----------
    paternity_probs: array
        An array of probabilties of paternity. This is usually supplied by a call
        to the `prob_array` function of a paternityArray object.

    partitions: array-like
        1-d array of integers labelling individuals into families. This should
        have as many elements as there are individuals in paternity_probs.

    includes_null: bool
        If True, the final column of paternity_probs is treated as containing
        the posterior probabilities that each individual is sired by an unsampled
        father. Defaults to True.

    ndraws: int
        The number of paths through fathers to be drawn. Set to 1000 by default.
        If you find that partitions have log likelihoods with differences less
        than 1 log likelihood unit apart you might want to increase this to get
        a more accurate estimate of the likelihood.

    Returns
    -------
    A float indicating log likelihood for the proposed partition.
    """
    # check the partition is the correct length!
    if paternity_probs.shape[0] != len(partitions):
        raise ValueError("paternity_probs and partition are of uneven length.")

    siblik = squash_siblings(paternity_probs, partitions) # multiply likelihoods within sibships
    # If all offspring go into a single partition
    if len(siblik.shape) is 1:
        return alogsumexp(siblik) # log likelihood of the model

    # If offspring go into different partitions.
    if len(siblik.shape) >  1:
        siblik = siblik[:, np.isfinite(siblik).sum(0) > 0] # remove columns with no non-zero terms
        prob_array = np.exp(siblik - alogsumexp(siblik, 1)[:, np.newaxis]) # normalise and exponentiate likelihoods

        # number of sibships and compatible fathers
        nfamilies = prob_array.shape[0]
        nfathers  = prob_array.shape[1]

        # If there are fewer possible fathers than sibships, then the likelihood must by definition be zero.
        if nfathers < nfamilies:
            with np.errstate(divide='ignore'):
                return np.log(0)

        else:
            # lists to loop over to save repeating them in the list comprehensions
            flist     = range(nfathers)
            slist     = range(nfamilies)

            # generate a sample of possible paths through the matrix of candidate fathers.
            path_samples = np.array([np.random.choice(flist, ndraws, replace=True, p = prob_array[i])
                                     for i in range(nfamilies)]).T
            path_samples = unique_rows(path_samples) # remove duplicate rows.

            # find paths which only visit each father once.
            # If unrelated fathers are not included in paternity_probs, remove all duplicates:
            if includes_null is False:
                unique_dads = [all(np.bincount(path_samples[i])      < 2) for i in range(len(path_samples))]
            # If unrelated fathers are included as the last column of paternity_probs, allow for multiple occurences of this.
            if includes_null is True:
                unique_dads = [all(np.bincount(path_samples[i]) < 2) and nfathers not in path_samples[i]
                               for i in range(len(path_samples))]
            # remove paths with shared fathers across two or more families.
            paths_to_keep   = np.arange(path_samples.shape[0])[np.array(unique_dads)] # indices for valid paths
            path_samples    = path_samples[paths_to_keep] # keep only valid paths

            # If no possible paths are found, return a likelihood of zero.
            if path_samples.shape[0] == 0:
                with np.errstate(divide='ignore'):
                    return np.log(0)
            # If there are possible paths, multiply over each set of fathers, and sum across sets.
            if path_samples.shape[0] > 0:
                # multiply over viable sets of fathers:
                path_product = [siblik[slist, path_samples[i]].sum() for i in range(path_samples.shape[0])]
                paths_sum    = alogsumexp(np.array(path_product)) # sum across sets
                return paths_sum



import numpy as np
from faps.genotypeArray import genotypeArray

def effective_nloci(progeny, mothers, adults):
    """
    Calculate the effective number of loci for all possible triplets in a comparison
    of offspring, known maternal, and candidate father genotypeArrays.

    Effective number of loci are the number of loci which for which all three
    individuals in the triplet have no missing genotype data.

    Parameters
    ----------
    progeny, mothers, adults: genotypeArray
        Genotype data on the progeny, their known mothers and the sample of candidate
        fathers

    Returns
    -------
    An array listing the number of valid loci for each possible triplet, with a row
    for each offspring and a column for each candidate father.
    """
    # confirm data are genotypeArrays
    if isinstance(progeny, genotypeArray) and isinstance(mothers, genotypeArray) and isinstance(adults, genotypeArray):
        if progeny.nloci != mothers.nloci and progeny.nloci != adults.nloci:
            raise ValueError("Numbers of loci in each genotypeArray do not match.")

        else:
            # Arrays indexing the positions of missing data in each genotypeArray
            prog_na = np.isnan(progeny.geno[:,:,0])[np.newaxis]
            moth_na = np.isnan(mothers.geno[:,:,0])[np.newaxis]
            adult_na= np.isnan(adults.geno[ :,:,0])[:,np.newaxis]
            # sum positions
            mat = prog_na + moth_na + adult_na
            # subtract from total for each triplet
            missing = progeny.nloci - (mat > 0).sum(2).T
            return missing
    else:
        raise TypeError("progeny, mothers and adults should be of type genotypeArray.")
        return None



from faps.genotypeArray import genotypeArray
from faps.alogsumexp import alogsumexp
import numpy as np
from warnings import warn

class paternityArray(object):
    """
    Likelihoods of that any of a set of candidate males is the true father of
    each offspring individual, assuming the mother is known. Call the wrapper
    function `paternity_array`.

    Parameters
    ----------
    likelihood: array
        Array of log likelihoods of paternity. Rows index offspring and columns
        candidates. The ij^th element is the likelihood that candidate j is the
        sire of offspring i.
    lik_absent: array
        Vector of log likelihoods that the true sire of each offspring is missing
        from the sample of candidate males, and hence that offspring alleles are
        drawn from population allele frequencies. Should be the same length as
        the number of rows in `likelihood`.
    offspring: array-like
        Identifiers for each offspring individual.
    mothers: array-like
        Identifiers of the mother of each individual, if these are known.
    fathers: array-like
        Identifiers of the father of each individual, if these are known.
    candidates: array-like
        Identifiers of the candidate fathers.
    mu: float
        Point estimate of the genotyping error rate. Note that sibship clustering
        is unstable if mu_input is set to exactly zero. Any zero values will
        therefore be set to a very small number close to zero (10^-12).
    missing_parents : float between zero and one
        Input value for the proportion of adults who are missing from the sample.
        This is used to weight the probabilties of paternity for each father
        relative to the probability that a father was not sampled. Defaults to
        zero, but you should definitely change this.
    purge: vector of str, or float between zero and one, optional
        Individuals who can be removed from the paternity array a priori. If
        a float is given, that proportion of individuals is removed from the
        array at random. Alternatively a string or vector of strings containing
        names of specific candidate fathers can be supplied.
    selfing_rate: float between zero and one, optional
        Input value for the prior probability of self-fertilisation.
    clashes: array, optional
        Array of the numbers of opposing homozygous incompatibilities for all
        possible parent-offspring dyads. This should have a row for every offspring
        and a column for every candidate father, and hence be the same shape as
        the `likelihood` array.
    max_clashes: int, optional
        Maximum number of double-homozygous incompatibilities allowed between
        offspring and candidate fathers. Dyads with more incompatibilities than
        this will have their probability set to zero.

    Returns
    -------
    prob_array: array
        Array of probabilities of paternity, accounting for the probability that
        a sire has not been sampled. This is the array `likelihood` with vector
        `lik_absent` appended, with rows normalised to sum to one.
    """

    def __init__(self, likelihood, lik_absent, offspring, mothers, fathers, candidates, missing_parents=0, mu=None, purge=None, selfing_rate=None, clashes = None, max_clashes = None, covariate=0):
        self.mu         = mu
        self.offspring  = offspring
        self.mothers    = mothers
        self.fathers    = fathers
        self.candidates = candidates
        self.lik_array  = likelihood
        self.lik_absent = lik_absent
        self.clashes    = clashes
        self.covariate  = covariate
        self.mu         = mu
        self.purge      = purge
        self.missing_parents = missing_parents
        self.selfing_rate = selfing_rate
        self.max_clashes = max_clashes

    def add_covariate(self, covariate):
        """
        Include a vector of (log) probabilities associated with covariates that
        provide additional information about paternity beyond that provided by
        genetic information (e.g. geographic distances).

        Parameters
        ----------
        covariate: 1-d array
            Vector of (log) probabilities of paternity based on non-genetic
            information, with one element for every candidate father. If this is a
            function of multiple sources they should be multiplied and included in
            this vector. If a list of offspring arrays have been supplied, this
            should be a list of vectors.

        Returns
        -------
        No output is printed; the covariate is added to the paternityArray as
        the attribute 'covariate'. Any existing information is overwritten. The
        vector is appended with an additional zero to allow for the final column
        of a the prob_array item in a paternityArray that accounts for the
        probability of missing fathers.
        """
        if isinstance(covariate, np.ndarray):
            if len(covariate.shape) > 1:
                raise ValueError("covariate should be a 1-d array, but has shape {}".format(covariate.shape))
            if len(self.candidates) != covariate.shape[0]:
                raise ValueError("Length of vector of covariates ({}) does not match the number of fathers ({})".format(len(self.candidates), covariate.shape[0]))
            if not all(covariate <= 0):
                warn("Not all values in covariate are less or equal to zero. Is it possible probabilities have not been log transformed?")
            covariate = np.append(covariate, 0)
            self.covariate = covariate
            return None
        else:
            raise TypeError("covariate should be a 1-d NumPy array.")


    def prob_array(self):
        """
        Construct an array of log posterior probabilities that each offspring is sired
        by each of the candidate males in the sample, or that the true father is not
        present in the sample. Rows are normalised to some to one.

        Additional arguments can specify the proportion of missing fathers, and the rate
        of self-fertilisation.

        Parameters
        ----------
        None.

        Returns
        -------
        An array with a row for each offspring individual and column for each
        candidate male, with an extra final column for the probability that the offspring
        is drawn from population allele frequencies. Each element is a log
        probability, and as such each row sums to one.

        Examples
        --------
        from faps import *
        import numpy as np

        # Generate a population of adults
        allele_freqs = np.random.uniform(0.3,0.5,50)
        adults = make_parents(20, allele_freqs)

        # Mate the first adult to the next three.
        mother = adults.subset(0)
        progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
        # Create paternityArray
        patlik = paternity_array(
            progeny,
            mother,
            adults,
            mu=0.0013,
            missing_parents=0.2
        )
        # View posterior probabilities of paternity (G matrix in the FAPS paper)
        patlik.prob_array()

        # Change some parameters that will affect the G matrix.
        patlik.missing_parents = 0.3
        patlik.selfing_rate = 0
        # Explicitly set some candidates to have probabilities of paternity to zero
        patlik.purge = 'base_0' # Another way to remove the mother
        patlik.purge = ['base_15', 'base_16', 'base_17'] # remove a list of specific candidates
        patlik.purge = 0.2 # Throw two candidates out at random

        # Another way to do the previous steps in a direct call to paternity_array.
        paternity_array(
            progeny,
            mother,
            adults,
            mu=0.0013,
            missing_parents=0.2,
            selfing_rate=0,
            purge = 0.2
        )
        """
        new_array = np.append(self.lik_array, self.lik_absent[:,np.newaxis], 1)

        # set log lik of individuals to be purged to -Inf
        if self.purge is not None:
            nc = len(self.candidates)
            # If a float is given, remove candidates at random.
            if isinstance(self.purge, float):
                if self.purge <= 0 or self.purge >= 1:
                    raise ValueError(" Error: purge must be between zero and one.")
                # Random set of candidate indices to be purged.
                ix = np.random.choice(range(nc), np.round(self.purge*nc).astype('int'), replace=False)
                with(np.errstate(divide='ignore')):
                    new_array[:, ix] = np.log(0)

            # If one or more integers is given, remove candidates at those indices
            elif isinstance(self.purge, (list, np.ndarray, int, str)):
                if isinstance(self.purge, (int, str)):
                    self.purge = [self.purge]
                # If all entries are strings, find the names of the candidates.
                if all([isinstance(x, str) for x in self.purge]):
                    if not all(np.isin(self.purge, self.candidates)):
                        raise ValueError("One or more names in paternityArray.purge are not found in paternityArray.candidates")
                    self.purge = [np.where(x == self.candidates)[0][0] for x in self.purge]
                with(np.errstate(divide='ignore')):
                    new_array[:, self.purge] = np.log(0)
            else:
                raise TypeError("Error: purge should be a float or list of floats between zero and one.")

        # correct for input parameter for proportion of missing fathers.
        if not isinstance(self.missing_parents, (int, float)):
            raise TypeError("missing_parents should be between 0 and 1.")
        # apply correction for the prior on number of missing parents.
        if self.missing_parents < 0 or self.missing_parents >1:
            raise ValueError("missing_parents must be between 0 and 1!")
        # if missing_parents is between zero and one, correct the likelihoods.
        if self.missing_parents >= 0 and self.missing_parents <= 1:
            if self.missing_parents ==0: warn("Missing_parents set to 0. Only continue if you are sure you really have 100% of possible fathers.")
            if self.missing_parents ==1: warn("Missing_parents set to 100%. Are you sure this is what you mean?")
            with(np.errstate(divide='ignore')):
                new_array[:, -1] = new_array[:, -1] + np.log(  self.missing_parents)
                new_array[:,:-1] = new_array[:,:-1] + np.log(1-self.missing_parents)
        # if missing_parents is 0, set the term for unrelated fathers to zero.
        if self.missing_parents == 0:
            with(np.errstate(divide='ignore')): new_array[:,-1] = np.log(0)

        # correct for selfing rate.
        if self.selfing_rate is not None:
            if not isinstance(self.selfing_rate, (int, float)):
                raise TypeError("selfing_rate should be between 0 and 1.")
            if self.selfing_rate < 0 or self.selfing_rate >1:
                raise ValueError("selfing_rate must be between 0 and 1.")
            if self.selfing_rate >=0 and self.selfing_rate <=1:
                if self.selfing_rate == 1: warn("selfing_rate set to 100%. Are you sure that is what you meant?")

                ix = range(len(self.offspring))
                with np.errstate(divide='ignore'):
                    maternal_pos = [np.where(np.array(self.candidates) == self.mothers[i])[0][0] for i in ix] # positions of the mothers
                    new_array[ix, maternal_pos] += np.log(self.selfing_rate)

        # set the likelihood dyads with many incompatibilities to zero
        if self.max_clashes is not None:
            if not isinstance(self.max_clashes, int):
                raise TypeError("paternityArray.max_clashes should be a positive integer")
            if self.max_clashes <= 0:
                raise ValueError("paternityArray.max_clashes should be greater than zero.")
            if self.clashes is None:
                raise TypeError("Unable to adjust for number of incompatible homozygous loci because `paternityArray.clashes` is `None`.")
            elif self.clashes.shape != self.lik_array.shape:
                raise ValueError("Shape of the likelihood array does not match that of the array of clashes.")
            else:
                inc = np.append(self.clashes, np.zeros(self.lik_absent.shape)[:,np.newaxis], 1) # add an extra column so the shapes match
                with np.errstate(divide='ignore'):
                    ix = np.log(inc <= self.max_clashes) # index elements to alter
                new_array = new_array + ix

        # normalise so rows sum to one.
        new_array = new_array - alogsumexp(new_array, axis=1)[:,np.newaxis]

        return new_array

    def subset(self, indices):
        """
        Subset offspring in a paternity array.

        Parameters
        ----------
        indices: List or array of integers
            Positions of individuals to subset.

        Returns
        -------
        A paternityArray object for the individuals indexed by `indices`.

        Examples
        --------
        from faps import *
        import numpy as np

        # Generate a population of adults
        allele_freqs = np.random.uniform(0.3,0.5,50)
        adults = make_parents(20, allele_freqs)

        # Mate the first adult to the next three.
        mother = adults.subset(0)
        progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
        # Create paternityArray
        patlik = paternity_array(progeny, mother, adults, mu=0.0013)

        # Pull out data for only the first family
        patlik.subset([0,1,2,3,4])
        """
        # If index is for a single individual, make it a list anyway.
        if isinstance(indices, int):
            individuals = [indices]
        # Subset original data.
        new_array = paternityArray(
            likelihood = self.lik_array[indices],
            lik_absent = self.lik_absent[indices],
            offspring  = self.offspring[indices],
            mothers    = self.mothers[indices],
            fathers    = self.fathers[indices],
            candidates = self.candidates
        )
        # Add additional attributes is these exist
        if self.mu           is not None: new_array.clashes      = self.mu
        if self.clashes      is not None: new_array.clashes      = self.clashes
        if self.covariate    is not None: new_array.covariate    = self.covariate

        # Return substted paternityArray
        return new_array

    def split(self, by, return_dict=True):
        """
        Split up a paternityArray into groups according to some grouping
        factor. For example, divide an array for multiple half-sibling
        arrays by the ID of their mothers.

        Parameters
        ----------
        by: array-like
            Vector containing grouping labels for each individual.
        return_dict: logical
            If True, the output is returned as a dictionary of paternityArray
            objects indexed by entries in `by`. If False, a list is returned.
            Defaults to True.

        Returns
        -------
        A dictionary of paternityArray objects.

        Examples
        --------
        from faps import *
        import numpy as np

        # Generate a population of adults
        allele_freqs = np.random.uniform(0.3,0.5,50)
        adults = make_parents(20, allele_freqs)

        # Mate the first adult to the next three.
        mother = adults.subset(0)
        progeny = make_sibships(adults, 0, [1,2,3], 5, 'x')
        # Create paternityArray
        patlik = paternity_array(progeny, mother, adults, mu=0.0013)

        # Pull out data for only the first family
        patlik.split(by=progeny.fathers)
        """
        groups = np.unique(by)
        ix = [np.where(by == i)[0] for i in groups]
        # Split into separate arrays.
        if return_dict:
            output = {k:self.subset(i) for k,i in zip(groups, ix)}
        else:
            output = [self.subset(i) for i in ix]
        #return output
        return output

    def write(self, path, decimals=3):
        """
        Write a matrix of (unnormalised) likelihoods of paternity to disk.

        Parameters
        ----------
        path: str
            Path to write to.
        decimals: int
            Number of decimal places to be saved to disk for likleihood values.

        Returns
        -------
        A CSV file indexing offspring ID, mother ID, followed by a matrix of likelihoods
        that each candidate male is the true father of each individual. The final
        column is the likelihood that the paternal alleles are drawn from population
        allele frequencies.
        """
        # append offspring and mother IDs onto the likelihood array.
        # append likelihoods of abset fathers on the back.
        newdata = np.append(self.offspring[:,np.newaxis],
                        np.append(self.mothers[:,np.newaxis],
                                  np.append(np.around(self.lik_array, decimals),
                                            np.around(self.lik_absent[:,np.newaxis], decimals),
                                            1),1),1)
        # headers
        cn = ','.join(self.candidates )
        cn = 'offspringID,motherID,' + cn + ',missing_father'
        # write to disk
        np.savetxt(path, newdata, fmt='%s', delimiter=',', comments='', header=cn)



import numpy as np
from faps.alogsumexp import alogsumexp
from faps.unique_rows import unique_rows
from faps.paternityArray import paternityArray
from faps.sibshipCluster import sibshipCluster
from faps.sibship_partitions import sibship_partitions
from warnings import warn

def do_clustering(paternity_array, ndraws=1000, exp=False, use_covariates=False):
    """
    Cluster offspring into full sibship groups using hierarchical clustering.

    This first builds a dendrogram of relatedness between individuals, and pulls
    out every possible partition structure compatible with the dendrogram. The
    likelihood for each partition is also estimated by Monte Carlo simulations.

    If the paternityArray object contains only a single individual, clustering
    is meaningless, but an artifical sibshipCluster object is created with a
    single partition with log likelihood of 0.

    Parameters
    ----------
    paternity_array: paternityArray
        Object listing information on paternity of individuals.
    ndraws: int
        Number of Monte Carlo simulations to run for each partition.
    exp: logical, optional
        Indicates whether the probabilities of paternity should be exponentiated
        before calculating pairwise probabilities of sibships. This gives a
        speed boost if this is to be repeated many times in simulations, but
        there may be a cost to accuracy. Defaults to False for this reason.
    use_covariates: logical, optional
        If True, information on prbabilities associated with covariates stored
        in paternityArray objects are incorporated into sibship clustering.

    Returns
    -------
    A sibshipCluster object.

    Example
    -------
    import numpy as np
    from faps import *

    # generate 4 families of 5 offspring
    allele_freqs = np.random.uniform(0.3, 0.5, 50)
    males = make_parents(200, allele_freqs)
    offspring = make_sibships(males, 0, range(1,5), 5)

    # Add muations
    mu = 0.0013
    males = males.dropouts(0.01).mutations(mu)
    offspring= offspring.dropouts(0.01).mutations(mu)

    mothers = males.subset(offspring.parent_index('m', males.names))
    patlik = paternity_array(offspring, mothers, males, mu)

    # compare the effect of exponentiating likelihoods
    sc1 = do_clustering(patlik, exp=True)
    sc2 = do_clustering(patlik, exp=False)

    sc1.accuracy(offspring, males)
    sc2.accuracy(offspring, males)
    """
    if not isinstance(paternity_array, paternityArray):
        TypeError("paternity_array should be a paternityArray object.")

    if use_covariates is True:
        if paternity_array.covariate is None:
            covar = 0
        elif isinstance(paternity_array.covariate, np.ndarray):
            if len(paternity_array.covariate.shape) > 1:
                raise ValueError("covariate should be a 1-d array, but has shape {}".format(covariate.shape))
            if paternity_array.prob_array().shape[1] != len(paternity_array.covariate):
                raise ValueError("Length of vector of covariates ({}) does not match the number of fathers ({})".format(len(paternity_array.covariate), paternity_array.prob_array().shape[1]))
            if not all(paternity_array.covariate <= 0):
                warn("Not all values in covariate are less or equal to zero. Is it possible probabilities have not been log transformed?")
            covar = paternity_array.covariate[np.newaxis]
        else:
            raise TypeError("If use_covariates is set to True, the covraiate attribute of the paternityArray object should be a 1-d array.")
    else:
        covar = 0

    # # If there is only one offspring in the paternityArray, there is no need to do any clustering.
    # if paternity_array.offspring.shape[0] == 0:
    #     output = sibshipCluster(
    #         paternity_array = paternity_array,
    #         linkage_matrix  = 'Linkage matrix not available because there is only one offspring',
    #         partitions      = np.array([[0]]),
    #         lik_partitions  = np.array([[0]]),
    #         paths           = {0: np.array([[0]])},
    #         path_likelihoods = {0: np.array([[0]])}
    #     )
    #     return output

    # # Otherwise, generate a sample of plausible partitions from the dendrogram.
    # else:
    partition_sample, z = sibship_partitions(paternity_array, exp=exp, use_covariates=use_covariates)

    #get a list of all unique families.
    unique_families = [np.where(p == val)[0] for p in partition_sample for val in np.unique(p)]
    unique_families = set([tuple(up) for up in unique_families]) # coerce to tuple to create set
    unique_families = [np.array(up) for up in unique_families] # coerce back to numpy arrays

    # Calculate likelihoods of paternity for each family.
    siblik = np.array([paternity_array.prob_array()[uf].sum(0) for uf in unique_families])
    siblik = siblik + covar
    sibprob= siblik - alogsumexp(siblik, 1)[:, np.newaxis]
    sibprob= np.exp(sibprob)

    # Draw ndraws fathers for each full sibship.
    candidates = np.arange(paternity_array.prob_array().shape[1])
    draw_dads = np.array([np.random.choice(candidates, ndraws, replace=True, p=s) for s in sibprob], dtype=np.int64)

    # Loop over each partition and pull out the rows in siblik for each full sibship within it.
    # Then, remove duplicate and invalid configurations.
    # Finally, extract probabilities for each candidate, multiply across families, and sum across valid paths.
    up_str = np.array([str(up) for up in unique_families])
    lik_partitions = np.zeros(len(partition_sample))
    paths = {}
    path_likelihoods = {}

    for i in range(len(partition_sample)):
        p = partition_sample[i]
        these_families = np.array([str(np.where(p == val)[0]) for val in np.unique(p)]) # families in this partition.
        # pull out the sublists in up_str for this partition
        ix = np.array([np.where(f == up_str)[0][0] for f in these_families])
        these_dads = draw_dads[ix]
        # remove invalid partitions
        these_dads = unique_rows(these_dads.T) # remove duplicates
        these_dads = np.array([v for v in these_dads if len(set(v)) == len(v)]) # remove partitions with duplicate candidates
        paths[i] = these_dads
        # If no valid partitions are found, return likelihood of zero.
        if len(these_dads) == 0:
            with np.errstate(divide='ignore'):
                lik_partitions[i] = np.array([np.log(0)])
                path_likelihoods[i] =  np.array([np.log(0)])
        # Otherwise, pull out the likelihoods
        else:
            this_path_lik       = siblik[ix, these_dads].sum(1)
            path_likelihoods[i] = this_path_lik
            lik_partitions[i]    = alogsumexp(this_path_lik)

    # Normalise likelihoods of individual paths over all partitions to sum to one.
    total_lik = alogsumexp(lik_partitions)[np.newaxis]
    path_probs = {k: v - total_lik for k,v in path_likelihoods.items()}

    output = sibshipCluster(
        paternity_array  = paternity_array,
        linkage_matrix   = z,
        partitions       = partition_sample,
        lik_partitions   = lik_partitions,
        paths            = paths,
        path_likelihoods = path_likelihoods,
        path_probs       = path_probs
    )
    return output



import numpy as np
from faps.paternityArray import paternityArray
from faps.sibshipCluster import sibshipCluster
from faps.matingEvents import matingEvents


def mating_events(sibship_clusters, paternity_arrays, family_draws = 1000, total_draws  = 10000, n_subsamples = 0, subsample_size = None, null_probs = None, family_weights=None, use_covariates=False, return_names=True):
    """
    Sample plausible mating events for multiple half-sibling arrays in parallel.

    This creates a single matingEvents object, where every half-sib family
    represents a unit, from which n=family_draws likely fathers are sampled.
    By default, the contribution of each half-sib family to the total sample
    is determined by the expected number of full-sib families in each array, but
    this can be changed by modifying family_weights.

    This is essentially a wrapper to perform sibshipCluster.mating_events for
    multiple half-sibling arrays at once. See the documentation for that function
    for details on sampling individual families.

    Parameters
    ----------
    sibship_clusters: dict
        Dictionary of sibshipCluster objects for multiple maternal families.
    paternity_arrays: dict
        Dictionary of paternityArray objects for multiple maternal families.
        Keys should match those for sibship_clusters.
    family_draws: int
        Number of mating events to sample for each maternal family.
    total_draws: int
        Total number of mating events to draw across maternal families.
    n_subsamples: int, optional
        Number of subsamples to draw from the total mating events.
    subsample_size: int, optional
        Number of mating events in each subsample. Defaults to 0.1*total_draws.
    null_probs: vector or dict, optional
        Array of (log) probabilities for paternity if this were not based on marker
        data. If the same probabilities apply to each half-sib array, this can
        be a vector with an element for each candidate father. Alternatively, if
        the vector for each array is different, supply an dictionary of 1-d
        vectors, with keys that match those for sibship_clusters. Probabilities
        in a given vector should sum to one.
    family_weights: array
        1-dimensional vector of weights to give to each half-sib array. If None,
        weights will be calculated as the mean number of full-sibships in each
        maternal family over possible partitions, weighted by the probability
        of each partition.
    return_names: logical, optional
        If True, unit_events and total_events are returned as the names of
        the candidates drawn for each maternal family. If False, their integer
        positions in the list of candidates is returned.

    Returns
    -------
    A matingEvents object. If null_probs is supplied samples for null mating are
    returned.

    See also
    --------
    sibshipCluster.mating_events()
    matingEvents()
    """
    if not isinstance(sibship_clusters, dict) or not all([isinstance(x, sibshipCluster) for x in sibship_clusters.values()]):
        raise TypeError('sibship_clusters should be a dictionary of sibshipCluster objects.')
    if not isinstance(paternity_arrays, dict) or not all([isinstance(x, paternityArray) for x in paternity_arrays.values()]):
        raise TypeError('paternity should be a dictionary of paternityArray objects.')
    elif len(sibship_clusters) != len(paternity_arrays):
        raise ValueError('Lists of sibshipCluster and paternityArray objects are different lengths.')
    elif len(sibship_clusters) == 1:
        warnings.warn('Lists of sibshipCluster and paternityArray are of length 1. If there is only one array to analyse it is cleaner to call mating_events dircetly from the sibshipCluster object.')

    # labels for each array, and each candidate
    candidates = list(paternity_arrays.values())[0].candidates
    candidates = np.append(candidates, 'father_missing')

    # Determine how to draw the correct number of samples for each half-sib array.
    # For a user-supplied vector of weights check this is the right format
    if isinstance(family_weights, np.ndarray):
        if len(family_weights.shape) > 1:
            raise ValueError("family_weights should be a 1-d vector, but has shape {}.".format(family_weights.shape))
        if len(family_weights) != len(sibship_clusters):
            raise ValueError('family_weights has length {}, but there are {} half-sib arrays sibship_clusters.'.format(len(family_weights), len(sibship_clusters)))
            family_weights = family_weights / family_weights.sum()
    # If no vector is given, use the expected mean number of full sibships per
    # maternal family.
    elif family_weights is None:
        family_weights = {k:v.mean_nfamilies() for k,v in sibship_clusters.items()} # expected number of mating events for each array
        total_weight   = total_draws / np.array(list(family_weights.values())).sum() # correction factor to sum to total_draws
        family_weights = {k : v * total_weight for k,v in family_weights.items()} # normalise to sum to total_draws
        family_weights = {k : np.around(v).astype('int') for k,v in family_weights.items()}
    else:
        raise TypeError('family_weights is type {}. If supplied, it should be a NumPy array.'.format(type(family_weights)))

    # If nothing is given for null_probs
    if null_probs is None:
        null_probs = {k: None for k in paternity_arrays.keys()}
    # If a NumPy array is given, coerce to a dictionary
    elif isinstance(null_probs, np.ndarray):
        # For 1-d vectors.
        if(null_probs.ndim == 1):
            if len(null_probs != len(candidates)-1):
                ValueError("If null_probs is a 1-d vector it should have the same length as the number of candidate fathers.")
            null_probs = {k:null_probs for k in paternity_arrays.keys()}
        # Throw an error if there is more than 1 dimension.
        elif(null_probs.ndim > 1):
            ValueError("null_probs should be a 1-d array or a dictionary of 1-d arrays.")
    # If a dictionary is given for null_probs, check it.
    elif isinstance(null_probs, dict):
        if any([k not in paternity_arrays.keys() for k in null_probs.keys()]):
            ValueError("One or more entries in null_probs does not have a matching entry in paternity_arrays.")
        if any([v.ndim > 1 for v in null_probs.values()]):
            ValueError("One or more entries in null_probs has more than one dimension.")
        if any([len(v) != len(candidates)-1 for k,v in null_probs.items()]):
            ValueError("One or more entries in null_probs is a different length to the number of candidates")
        if not all([isinstance(v, np.ndarray) for k,v in null_probs.items()]):
            TypeError("If null_probs is a dictionary, all values should be 1-d NumPy arrays.")
    else:
        raise TypeError("null_probs is type {}.\nIf supplied, this should be a NumPy array, or dictionary.").format(type(null_probs))

    # draw events for each half-sib array
    family_events = {}
    total_events  = {}
    for id in sibship_clusters.keys():
        me = sibship_clusters[id].mating_events(
            paternity_array = paternity_arrays[id],
            unit_draws=family_draws,
            total_draws=total_draws,
            n_subsamples=n_subsamples,
            subsample_size=subsample_size,
            null_probs=null_probs[id],
            use_covariates=use_covariates,
            return_names = return_names
        )
        family_events[id] = me.total_events
        # Random subset of fathers from each maternal family, with sample size in
        # proportion to the weight of the family
        total_events[id] = np.random.choice(me.total_events, family_weights[id], replace=True)
    # Concatenate total_events into a single list.
    total_events = [item for sublist in list(total_events.values()) for item in sublist]
    total_events = np.array(total_events)

    # Create a matingEvents object.
    me = matingEvents(
        unit_names   = np.array(list(family_events.keys())),
        candidates   = candidates,
        unit_weights = family_weights,
        unit_events  = family_events,
        total_events = total_events
    )
    # draw subsamples.
    me.subsamples = me.draw_subsample(n_subsamples, subsample_size)


    return me



import numpy as np
from faps.unique_rows import unique_rows
import fastcluster
from scipy.cluster.hierarchy import fcluster
from faps.pairwise_lik_fullsibs import pairwise_lik_fullsibs

def sibship_partitions(paternity_array, exp=False, method='average', criterion = 'distance', use_covariates=True):
    """
    Generate a sample of partition structures from an array of likelihoods that
    each pair of individuals in a half-sib array are really full sibs.

    This performs hierarchical clustering using the fastcluster and scipy
    libraries. The default algorithm is UPGMA, but linkage and clustering
    functions can be tweaked using the method and criterion functions.

    Parameters
    ----------
    paternity_array: paternityArray
        Object listing information on paternity of individuals.
    method: str
        Distance function passed to linkage. See fastcluster.linkage for
        available inputs. Defaults to 'average'.
    criterion: str
        Clustering function used to bisect the dendrogram. See
        scipy.hierarchy.fcluster for available inputs. Defaults to 'distance'.
    use_covariates: logical, optional
        If True, information on prbabilities associated with covariates stored
        in paternityArray objects are incorporated into sibship clustering.
        Defaults to True

    Returns
    -------
    An array of plausible partitions that group individuals into full sibships.
    In each row of the output individuals are labelled with an integer that
    groups them in a full sibship with all other individuals with that label.

    Example
    -------
    import numpy as np
    from faps import *

    # generate 4 families of 5 offspring
    allele_freqs = np.random.uniform(0.3, 0.5, 50)
    males = make_parents(200, allele_freqs)
    offspring = make_sibships(males, 0, range(1,5), 5)

    # Add muations
    mu = 0.0013
    males = males.dropouts(0.01).mutations(mu)
    offspring= offspring.dropouts(0.01).mutations(mu)

    mothers = males.subset(offspring.parent_index('m', males.names))
    patlik = paternity_array(offspring, mothers, males, allele_freqs, mu)

    # Generate possible sets of partitions
    partition_sample = sibship_partitions(patlik)
    """
    if len(paternity_array.offspring) == 1:
        partition_sample = np.array([[0]])
        z = 'Linkage matrix not available because there is only one offspring'

    else:
        fullpairs = pairwise_lik_fullsibs(paternity_array, use_covariates=use_covariates, exp = exp)
        # Clustering matrix z.
        z= fastcluster.linkage(abs(fullpairs[np.triu_indices(fullpairs.shape[0], 1)]), method)
        z = np.clip(z, 0, 10**12)
        # A list of thresholds to slice through the dendrogram
        thresholds = np.append(0,z[:,2])
        # store all possible partitions from the dendrogram
        partition_sample = [fcluster(z, t, criterion) for t in thresholds]
        partition_sample = unique_rows(partition_sample)

    return partition_sample, z



import numpy as np
from faps import make_offspring

def make_sibships(parents, dam, sires, family_size, family_name='offs'):
    """
    Mate parents in a base population to create half- or full-sibling families.

    This relies on the indices for the desired parents within the genotype object
    'parents'. These can be found from genotype objects using the function
    parent_index().

    Parameters
    ----------
    parents: genotypeArray
        Genotype information on the parents to be mated.

    dam: int
        Position of the mother in the parental genotypeArray object, to whom
        multiple sires are mated.

    sires: list
        A list indexing the position of the sires to be mated to the dam(s)
        within the parental genotype object. To create a full sibship, supply
        only a single father.

    family_size: int, list
        The sizes of each full sibship. If famillies are to be of the
        same size give and integer. Alternatively, if different sizes are
        desired, supply a list of sizes for each parternal familes. This list
        must be the same length as the list of sires.

    family_name: str
        String denoting the name for this family.

    Returns
    -------
    A genotypeArray object.
    """
    # if there is only one sire, turn the integer into a list of length one.
    if isinstance(sires, int):
        sires = [sires]

    # Multiply each sire ID by the number of his offspring.
    if isinstance(family_size, int): # for equal family sizes.
        sire_list = np.sort(sires*family_size).tolist()

    # if family_size is given as a list
    if isinstance(family_size, list) or isinstance(family_size, np.ndarray):
        # return an error if the two lists are of unequal length.
        if len(family_size) != len(sires):
            raise ValueError("family_size must either be an integer, or a list of the same length as the list of sires.")
        else:
            sire_list = [[sires[x]] * family_size[x] for x in range(len(sires))]
            sire_list = [item for sublist in sire_list for item in sublist]

    # Replicate the dam ID to match the length of sire_list
    dam_list = [dam] * len(sire_list)

    genotypes = make_offspring(parents, dam_list=dam_list, sire_list=sire_list, family_name=family_name)

    return genotypes



import numpy as np
from faps.paternityArray import paternityArray
from faps.alogsumexp import alogsumexp

def pairwise_lik_fullsibs(paternity_array, use_covariates = False, exp = False):
    """
    Create a matrix of probabilities that each pair of individuals in a
    half-sibling array are full siblings.

    If two individuals are full siblings, the likelihood is the product of the
    likelihoods that each is fathered by father i, or father j, etc. We therefore
    take the product of likelihoods for paternity for each candidtae, and sum
    across the products for each father to account for uncertainty.

    The majority of fathers will not contribute to paternity, but will cause
    noise in the estimation of pairwise sibship probability. For this reason the
    function works much better if the likelihood array has had most unlikely
    fathers, for example by setting the likelihood for fathers with three or
    more opposing homozygous loci to zero.

    Parameters
    ----------
    paternity_array: paternityArray
        Object listing information on paternity of individuals.
    exp: bool, optional
        If True, probabilities of paternity are exponentiated before calculating
        pairwise probabilities of sibships. This gives a speed boost if this is
        to be repeated many times, but there may be a cost to accuracy (this is
        untested!).
    use_covariates: logical, optional
        If True, information on prbabilities associated with covariates stored
        in paternityArray objects are incorporated into sibship clustering.

    Returns
    -------
    A noffspring*noffspring matrix of log likelihoods of
    being full siblings.

    Example
    -------
    import numpy as np
    from faps import *

    # generate 4 families of 5 offspring
    allele_freqs = np.random.uniform(0.3, 0.5, 50)
    males = make_parents(200, allele_freqs)
    offspring = make_sibships(males, 0, range(1,5), 5)

    # Add muations
    mu = 0.0013
    males = males.dropouts(0.01).mutations(mu)
    offspring= offspring.dropouts(0.01).mutations(mu)

    mothers = males.subset(offspring.parent_index('m', males.names))
    patlik = paternity_array(offspring, mothers, males, allele_freqs, mu)

    # Matrix of pairwise probabilities of being full siblings.
    fullpairs = pairwise_lik_fullsibs(patlik)
    """
    paternity_probs = paternity_array.prob_array()
    # pull out the number of offspring and parents
    noffs     = paternity_probs.shape[0]
    nparents  = paternity_probs.shape[1]

    if exp is False:
        # Set up use of covariates
        if use_covariates is True:
            covar = paternity_array.covariate[np.newaxis, :, np.newaxis]
        else: # If they are not to be used, use a log prob of 0.
            covar = 0

        # this section of code calculates the matrix in log space, but I found it quicker to exponentiate (below).
        paternity_probs = paternity_probs[:,:, np.newaxis]
        # take all pairwise products of sharing fathers.
        pairwise_lik = np.array([alogsumexp(i + paternity_probs.T + covar, 1) for i in paternity_probs]).squeeze()

        return pairwise_lik
    if exp is True:
        # Set up use of covariates
        if use_covariates is True:
            covar = paternity_array.covariate[np.newaxis]
        else: # If they are not to be used, use a log prob of 0.
            covar = 0

        # the sum can be quicker if you exponentiate, but this may harm precision.
        paternity_probs = np.exp(paternity_probs)
        covar = np.exp(covar)
        # for each pair of offspring, the likelihood of not sharing each father.
        pairwise_lik = np.matmul(paternity_probs*covar, paternity_probs.T)
        pairwise_lik = np.array(pairwise_lik).reshape([noffs, noffs]) # reshape

        return np.log(pairwise_lik)



import numpy as np

def calculate_geno_probs(geno, mu):
    """
    Calcuate probabilities that each individual has genotype 0, 1 or 2.

    Parameters
    ----------
    mu: float or 1-d array between 0 and 1
        Per locus genotype error rate; the probability that the called
        genotype is incorrect. Alternatively, supply a vector of error rates
        for each locus.

    Returns
    -------
    A 3-d array of probabilities that each observed marker data point is 0,
    1 or 2 given the input error rate(s).
    Axes index:
    1. Possible genotypes 0, 1 or 2
    2. Individuals
    3. Loci

    Examples
    --------
    from faps import *
    import numpy as np

    # Generate a population of adults
    allele_freqs = np.random.uniform(0.3,0.5,3)
    adults = make_parents(20, allele_freqs)

    # Example with a single error rate
    adults.geno_probs(mu = 0.01)
    # Example with a vector of errors
    adults.geno_probs(mu = np.array([0.01, 0.02, 0.03]))
    """
    if isinstance(mu, float):
        if np.array([mu < 0, mu >1]).any():
            raise ValueError("mu should be between 0 and 1")
    elif isinstance(mu, np.ndarray):
        if (len(mu.shape) != 1):
            raise ValueError("mu should be a 1-dimensional vector or float.")
        if np.array([mu < 0, mu >1]).any():
            raise ValueError("All elements in mu should be between 0 and 1")
        if len(mu) != geno.shape[1]:
            raise ValueError("mu should either be a single float or have the same number of elements as the number of loci.")
    else:
        raise TypeError("mu should either be a float or vector of floats between 0 and 1.")

    diploid = geno.sum(2)
    # 3-d array giving positions that are 0, 1 or 2.
    probs = np.array([diploid == i for i in [0,1,2]])
    # Transpose to index indivudals, loci, and possible genotype
    probs = probs.transpose([1,2,0])
    # Insert mu where probs is True, and mu/2 if False
    # mu/2 because we don't know which other genotype it could be.
    probs = (probs * (1-mu)) + (~probs * mu/2)
    # Where data are not observed, we set this to NaN.
    # probs[diploid < 0] = np.nan

    return probs
